"""Storpool cgconf."""

from __future__ import print_function, division
import argparse
import collections
import os
import sys
import textwrap


from sp import cpuinfo
from sp.cgroups import hierarchy, filters
from sp import messages as m
from sp.util import limits as limutils
from sp.util.file.writers import write_file_with_backup, write_file
from sp.util.misc import get_memtotal_kb

from .cgtool import config, spcpu, migration, spmem, cgtemplates
from .cgtool.config import generate as confgen
from .cgtool.config import MAX_SERVERS
from .cgtool.util import spcg
from .cgtool.util.print import print_mem_conf, print_cpu_conf


OPTIONS = collections.OrderedDict([
    ('CONVERGED', '(boolean) specify whether the machine is hyperconverged'),
    ('BLOCK', '(boolean) specify whether the block service is installed'),
    ('ISCSI', '(boolean) specify whether the iscsi service is installed'),
    ('MGMT', '(boolean) specify whether the mgmt service is installed'),
    ('BRIDGE', '(boolean) specify whether the bridge service is installed'),
    ('SERVERS', '(integer) number of StorPool servers'),
    (
        'SET_CACHE_SIZE',
        '(boolean) specify whether to calculate and set new server cache size'
    ),
    (
        'SET_MEMSW',
        '(boolean) specify whether to set memory.memsw.limit in the cgconfig '
        'file'
    ),
    (
        'SYSTEM_LIMIT',
        'specify the memory limit for the system slice (default: 2G)'
    ),
    (
        'USER_LIMIT',
        'specify the memory limit for the user slice (default: 2G)'
    ),
    ('MACHINE_LIMIT', 'specify the memory limit for the machine slice'),
    ('KERNEL_MEM', 'memory that will be left to the kernel (default: 1G)'),
    (
        'SP_COMMON_LIMIT',
        'specify the memory limit for the storpool/common slice'
    ),
    (
        'SP_ALLOC_LIMIT',
        'specify the memory limit for the storpool/alloc slice'
    ),
    ('IFACE', 'specify the network interface for StorPool'),
    (
        'IFACE_ACC',
        '(boolean) specify whether there is/isn\'t interface acceleration'
    ),
    ('CORES', 'number of cores dedicated for StorPool'),
    ('NUMA_OVERFLOW', 'allow the usage of non-iface-local cpus for StorPool'),
    ('CONFDIR', 'specify the directory to write the config files to'),
    ('CACHEDIR', 'specify the directory to write the cache config to'),
])


def get_options_help():
    """Return cgconf options help string."""
    line_max_chars = 80 - 2
    option_name_chars = max(map(len, OPTIONS.keys()))

    def wrap(descr):
        """Wrap (split) the line to fit in the desired char length."""
        sep = '\n' + (option_name_chars + 1) * ' '
        max_descr_chars = line_max_chars - option_name_chars - 1 - 3 * 8
        return sep.join(textwrap.wrap(descr, max_descr_chars))

    opt_tmpl = '{0:' + str(option_name_chars) + 's} {1}'
    return '\n'.join(opt_tmpl.format(k, wrap(v)) for (k, v) in OPTIONS.items())


def valid_file(filep):
    """Filename parser/validator for argparse.ArgumentParser."""
    if not os.path.isfile(filep):
        raise argparse.ArgumentTypeError("Invalid file {0}".format(filep))
    return filep


def key_equals_value(kvstr):
    """Key-value string pair parser/validator for argparse.ArgumentParser."""
    kvpair = kvstr.split('=')
    if len(kvpair) != 2:
        raise argparse.ArgumentTypeError(
            "Invalid config pair {0}".format(kvstr)
        )
    return tuple(kvpair)


def get_args_parser():
    """Return command line arguments parser."""
    parser = argparse.ArgumentParser(
        description='StorPool control group configuration',
        formatter_class=argparse.RawTextHelpFormatter,
        prog='storpool_cg conf'
    )
    parser.add_argument('-N', '--noop', action='store_true', default=False,
                        help='No operation, only print the config.')
    parser.add_argument('-M', '--migrate', action='store_true', default=False,
                        help='Apply the configuration live.')
    parser.add_argument('-L', '--preserve-limits', action='store_true',
                        default=False,
                        help='Preserve existing memory limits when migrating')
    parser.add_argument('-P', '--preserve-cpus', action='store_true',
                        default=False,
                        help='Preserve cpus and cpu bindings when migrating')
    parser.add_argument('-C', '--configfile', type=valid_file,
                        help="Configuration file for the tool.")
    parser.add_argument('-D', '--dump-configfile', type=str,
                        help='Write the config for the tool in the file')
    parser.add_argument('config', nargs='*', type=key_equals_value,
                        help='Optional overrides of the config [key=value]'
                        + '\n' + get_options_help())
    parser.add_argument('-v', '--verbose', action='count', default=0,
                        help='Verbose level')
    parser.add_argument('-i', '--interactive', action='store_true',
                        default=False,
                        help='Migration interactive mode')
    parser.add_argument('-E', '--expand', action='store_true', default=False,
                        help='Allow the migration to increase the cpus count '
                             'for StorPool')
    return parser


def get_args():
    """Return the parsed command line arguments."""
    parser = get_args_parser()
    args = parser.parse_args()
    if args.interactive and not args.migrate:
        parser.error('cannot use --interactive without --migrate')
    if args.expand and not args.migrate:
        parser.error('cannot use --expand without --migrate')
    return args


def check_cpu():
    """Check if the machine processors are not supported."""
    cpui = cpuinfo.get()
    if cpui.vendor() == 'AuthenticAMD' and cpui.family() <= 22:
        m.exit_msg('AMD family {f} cpus are not supported by the cgtool.'
                   .format(f=cpui.family()))


def get_slices_info(cfg, sp_cpus):
    """Generate slices config for cgconfigparser."""
    non_sp_cpus = sorted(set(cpuinfo.get().list_all_cpus()) - set(sp_cpus))
    cpuset_info = hierarchy.get().cpuset_info()
    slices_info = {}
    for slc, limit in cfg.memconf.get_limits().items():
        slc_name = '{0}.slice'.format(slc)
        cpuset = sp_cpus if slc == 'storpool' else non_sp_cpus
        if cfg.optconf.migrate and slc_name in cpuset_info:
            cpuset = cpuset_info[slc_name]['cpuset.cpus']
        slices_info[slc_name] = {'limit': limit, 'cpuset': cpuset}

    # warn for non-cgtool slices cpulists
    for slc_name, info in hierarchy.get().cpuset_info().items():
        if not filters.is_direct_root_child(slc_name):
            continue
        if slc_name not in slices_info:
            cpuset = str(info['cpuset.cpus'])
            m.warn_msg('{slc} is not a cgtool managed slice. Plase make sure '
                       'that the configuration file for it has the following '
                       'cpuset: {cpus}'.format(slc=slc_name, cpus=cpuset))

    return slices_info


def write_conf(cfg, sp_cpulist, s_to_cpu):
    """Write the cgroups configuration in the specified directory."""
    slices_info = get_slices_info(cfg, sp_cpulist)
    slices_strs = cgtemplates.gen_slices_strs(slices_info, s_to_cpu,
                                              cfg.memconf.set_memsw)
    for slice_name, slice_conf in slices_strs.items():
        fpath = os.path.join(cfg.dirconf.confdir, slice_name + '.conf')
        backup_path = os.path.join(
            cfg.dirconf.confdir, 'old', slice_name + '.conf'
        )
        write_file_with_backup(fpath, slice_conf, backup_path)
    # mgmt backup
    mgmt_path = os.path.join(cfg.dirconf.confdir, 'mgmt.slice.conf')
    mgmt_backup_path = os.path.join(
        cfg.dirconf.confdir, 'old', 'mgmt.slice.conf'
    )
    if os.path.isfile(mgmt_path):
        os.rename(mgmt_path, mgmt_backup_path)

    # machine.slice backup after migrate
    machine_path = os.path.join(cfg.dirconf.confdir, 'machine.slice.conf')
    machine_backup_path = os.path.join(
        cfg.dirconf.confdir, 'old', 'machine.slice.conf'
    )
    if 'machine.slice' not in slices_strs and os.path.isfile(machine_path):
        os.rename(machine_path, machine_backup_path)


def write_cache_conf(cfg):
    """Write the cache configuration in the specified directory."""
    cache_conf = spmem.get_cache_conf(cfg.memconf.set_cache_size) + '\n'
    cache_fpath = os.path.join(cfg.dirconf.cachedir, 'cache-size.conf')
    backup_path = os.path.join(cfg.dirconf.cachedir, 'old', 'cache-size.conf')
    write_file_with_backup(cache_fpath, cache_conf, backup_path)


def should_update_cache_first():
    """Check if storpool_server cache should be updated before migrating."""
    cur_cache = spmem.cache_sizes_from_conf()
    needed_cache, _ = spmem.servers_needs()
    return sum(cur_cache) > sum(needed_cache)


def cgmain(cfg):
    """Create/migrate to the desired cgroups configuration."""
    # pylint: disable=too-many-branches
    if cfg.optconf.preserve_cpus:
        cpu_info = hierarchy.get().cpuset_info()
        sp_cpulist = cpu_info['storpool.slice']['cpuset.cpus']
        cpu_to_s = spcg.current_cpu_to_service(cpu_info)
        s_to_cpu = spcpu.gen_service_to_cpu(cpu_to_s)
    else:
        s_to_cpu, cpu_to_s = spcpu.assign_cpus(cfg)
        sp_cpulist = cpu_to_s.keys()

    if cfg.optconf.noop:
        print('########## START SUMMARY ##########')
        print_mem_conf(cfg.memconf.get_limits())
        print('###################################')
        print("cpus for StorPool: {0}".format(sorted(sp_cpulist)))
        print_cpu_conf(cpu_to_s)
        if cfg.memconf.set_cache_size:
            print('###################################')
            print(spmem.get_cache_conf(cfg.memconf.set_cache_size))
        print('########### END SUMMARY ###########')
        if cfg.optconf.migrate:
            if cfg.memconf.set_cache_size and should_update_cache_first():
                m.warn_msg('The new server cache config has been written. '
                           'Please run update_rdma first and start the '
                           'migration again with set_cache_size=0')
                return

            hier = hierarchy.get(mock=True)
            migation_ok = migration.apply_config(
                cfg, hier, sp_cpulist, s_to_cpu
            )
            if migation_ok and cfg.memconf.set_cache_size and \
                    not should_update_cache_first():
                m.warn_msg('The new server cache config has been written. '
                           'Remember to restart the servers.')
    else:
        if cfg.optconf.migrate:
            if cfg.memconf.set_cache_size and should_update_cache_first():
                write_cache_conf(cfg)
                m.warn_msg('The new server cache config has been written. '
                           'Please run update_rdma first and start the '
                           'migration again with set_cache_size=0')
                return

            hier = hierarchy.get()
            try:
                should_write_conf = migration.apply_config(
                    cfg, hier, sp_cpulist, s_to_cpu
                )
            except migration.MigrationError as err:
                m.err_msg(err)
                should_write_conf = False
            if should_write_conf:
                write_conf(cfg, sp_cpulist, s_to_cpu)
                if cfg.memconf.set_cache_size and \
                        not should_update_cache_first():
                    write_cache_conf(cfg)
                    m.warn_msg('The new server cache config has been written. '
                               'Remember to restart the servers.')
        else:
            if cfg.memconf.set_cache_size:
                write_cache_conf(cfg)
            write_conf(cfg, sp_cpulist, s_to_cpu)


def get_overrides(args):
    """Get command line arguments overrides from current configuration."""
    overrides = dict(args.config)

    def add_limit_override(meminfo, slice_name, override_name):
        """Add memory limit override."""
        if slice_name in meminfo:
            mb_limit = meminfo[slice_name]['memory.limit_in_bytes'] / 1024**3
            overrides[override_name] = limutils.limit_to_gb_str(mb_limit)
            m.msg('Preserved {0} limit, all config options will be ignored.'
                  .format(slice_name))
        else:
            m.msg('Couldn\'t preserve {0} limit, because slice does not exist'
                  .format(slice_name))

    if args.preserve_limits:
        m.msg('Preserving memory limits. Will use set_cache_size=0')
        overrides['set_cache_size'] = '0'
        meminfo = hierarchy.get().memory_info()
        add_limit_override(meminfo, 'user.slice', 'user_limit')
        add_limit_override(meminfo, 'system.slice', 'system_limit')
        add_limit_override(meminfo, 'machine.slice', 'machine_limit')
        add_limit_override(meminfo, 'storpool.slice/common', 'sp_common_limit')
        add_limit_override(meminfo, 'storpool.slice/alloc', 'sp_alloc_limit')
        limit_sum_b = sum(
            meminfo[slc]['memory.limit_in_bytes']
            for slc in filter(filters.is_direct_root_child, meminfo)
        )
        kernel_mem_mb = (get_memtotal_kb() - (limit_sum_b / 1024)) / 1024
        kernel_mem_mb = max(0, kernel_mem_mb)  # avoid negative values
        m.msg('Preserving memory left for kernel({0}MB)'.format(kernel_mem_mb))
        overrides['kernel_mem'] = limutils.limit_to_gb_str(
            kernel_mem_mb / 1024
        )

    if args.preserve_cpus:
        cpu_info = hierarchy.get().cpuset_info()
        sp_cpus = len(cpu_info['storpool.slice']['cpuset.cpus'])
        sp_cores = sp_cpus / (1 + cpuinfo.get().hyperthreading())
        overrides['cores'] = str(sp_cores)
        s_to_cpu = spcpu.gen_service_to_cpu(
            spcg.current_cpu_to_service(cpu_info)
        )
        for service in ('block', 'mgmt', 'bridge', 'iscsi'):
            overrides[service] = '1' if service in s_to_cpu else '0'
            m.msg('Preserving cpusets. Will use {0}={1}'
                  .format(service, overrides[service]))
        overrides['servers'] = '1' if 'server' in s_to_cpu else '0'
        for i in range(1, MAX_SERVERS):
            if 'server_{0}'.format(i) in s_to_cpu:
                overrides['servers'] = str(i + 1)
        m.msg('Preserving cpusets. Will use servers={0}'
              .format(overrides['servers']))

    return overrides


def optconf_from_args(args):
    """Create OptionsConfig from the command line arguments."""
    optconf = config.OptionsConfig()
    optconf.noop = args.noop
    optconf.migrate = args.migrate
    optconf.interactive = args.interactive
    optconf.preserve_cpus = args.preserve_cpus
    optconf.preserve_limits = args.preserve_limits
    optconf.expand = args.expand
    return optconf


def get_cfg(args):
    """Create ControlGroupToolConfig from the command line arguments."""
    config_overrides = get_overrides(args)
    parser = config.parser.get_file_parser(args.configfile, config_overrides)
    cfg = config.ControlGroupToolConfig()
    cfg.optconf = optconf_from_args(args)
    cfg.dirconf = confgen.get_directory_conf(parser)
    cfg.netconf = confgen.get_network_conf(parser)
    cfg.serviceconf = confgen.get_services_conf(parser)
    cfg.memconf, cfg.typeconf = confgen.get_memory_and_type_confs(
        parser, cfg.optconf, cfg.serviceconf
    )
    cfg.cpuconf = confgen.get_cpu_conf(parser, cfg.serviceconf, cfg.netconf,
                                       cfg.typeconf)
    cfg.sanity_check()
    return cfg


def main(argv=None):
    """
    The main program: parse the command line arguments and create/migrate to
    the desired cgroups configuration.
    """
    if argv is not None:
        sys.argv = argv
    check_cpu()
    args = get_args()
    m.set_verbose_level(args.verbose)
    cfg = get_cfg(args)
    if args.dump_configfile:
        write_file(args.dump_configfile, str(cfg))
    cgmain(cfg)
