"""Storpool cgprint."""

from __future__ import print_function, division

import argparse
import sys

from sp import cpuinfo
from sp.cgroups import hierarchy, filters
from sp.util import limits as limutils
from sp.util import misc as utils

from .cgtool.util import spcg


def get_args_parser():
    """Return command line arguments parser."""
    parser = argparse.ArgumentParser(description='StorPool cgroups printing',
                                     prog='storpool_cg print')
    parser.add_argument('-N', '--numas', action='store_true', default=False,
                        help='Print the numa nodes in the cpu config.')
    parser.add_argument('-S', '--slices', action='store_true', default=False,
                        help='Print the cpuset slices in the cpu config.')
    return parser


def get_args():
    """Return the parsed command line arguments."""
    parser = get_args_parser()
    args = parser.parse_args()
    return args


def print_memory(memory_info):
    """Display the current cgroups memory configuration."""
    for grp, info in sorted(memory_info.items()):
        limit_gb = info['memory.limit_in_bytes'] / 1024**3
        limit = limutils.limit_to_gb_str(limit_gb)
        if filters.is_direct_root_child(grp):
            print('slice: {0} limit: {1}'.format(grp, limit))
        if filters.is_sp_subgroup(grp):
            print('  subslice: {0} limit: {1}'.format(grp, limit))


BASE_TMPL = '  core:{core} cpus:[{cpus}]'
NUMA_TMPL = '  --  numa:[{numas}]'
SLCS_TMPL = '  --  {slices}'


def print_cpu_info(args, cpu_info, cpu_to_s):
    """Display the current cgroups cpu configuration."""
    # pylint: disable=too-many-locals
    cpui = cpuinfo.get()

    def _thread_slices(thread):
        def _strip_slice(slc):
            return slc[:-6] if slc.endswith('.slice') else slc

        def _slice(slc):
            slc = _strip_slice(slc)
            if slc == 'storpool':
                services = ','.join(cpu_to_s.get(thread, []))
                slc = ('storpool: {0}'.format(services) if args.slices
                       else services)
            return slc

        slc_filter = (filters.is_direct_root_child if args.slices
                      else lambda slc: slc == 'storpool.slice')
        return sorted(_slice(slc) for slc, info in cpu_info.items()
                      if slc_filter(slc) and thread in info['cpuset.cpus'])

    infos = []
    socket_starts = {}
    for socket in cpui.socket_ids():
        socket_starts[len(infos)] = socket
        for core, threads in cpui.socket_info(socket).items():
            threads = sorted(threads)
            numas = [str(cpui.numa_node_of(t)) for t in threads]
            slices = [' '.join(_thread_slices(t)) for t in threads]
            infos.append((str(core), map(str, threads), numas, slices))

    def _col_lens(col):
        if utils.is_string(infos[0][col]):
            return max(len(info[col]) for info in infos)
        if isinstance(infos[0][col], list):
            return [max(len(info[col][i]) for info in infos)
                    for i in range(len(infos[0][col]))]
        raise TypeError()

    def _value_align(value):
        try:
            int(value)
            return '>'
        except ValueError:
            return '<'

    def _list_align(lst):
        return [_list_align(value) if isinstance(value, list)
                else _value_align(value) for value in lst]

    def _add_widths(info, lengths, alignments):

        def _add_format(col, width, align):
            if isinstance(width, int):
                return ('{0:' + align + str(width) + '}').format(col)
            if isinstance(width, list):
                assert isinstance(col, list) and isinstance(align, list)
                return _add_widths(col, width, align)
            raise TypeError()

        return map(_add_format, info, lengths, alignments)

    lengths = [_col_lens(col) for col in range(len(infos[0]))]
    alignments = _list_align(infos[0])

    tmpl = (BASE_TMPL + NUMA_TMPL + SLCS_TMPL if args.numas
            else BASE_TMPL + SLCS_TMPL)

    for line_number, info in enumerate(infos):
        if line_number in socket_starts:
            print("socket:{0}".format(socket_starts[line_number]))
        core, cpus, numas, slices = _add_widths(info, lengths, alignments)
        if all(map(str.isspace, slices)):
            slices = []
        print(tmpl.format(core=core,
                          cpus=' '.join(cpus),
                          numas=' '.join(numas),
                          slices=' | '.join(slices)))


def main(argv=None):
    """
    The main program: parse the command line arguments and display the cgroups
    configuration.
    """
    if argv is not None:
        sys.argv = argv
    args = get_args()

    cpu_info, memory_info = hierarchy.get().cpuset_memory_info()

    cpu_to_s = spcg.current_cpu_to_service(cpu_info)
    # add nic to first cpu, if it is empty
    if 'storpool.slice' in cpu_info:
        sp_cpuset = cpu_info['storpool.slice']['cpuset.cpus']
        possible_nic_cpus = [
            cpu for cpu in sp_cpuset
            if not cpuinfo.get().is_on_first_core_of_numa(cpu)
        ]
        if possible_nic_cpus:
            nic_cpu = min(possible_nic_cpus)
            cpu_to_s[nic_cpu] = cpu_to_s.get(nic_cpu, ['nic'])

    print_memory(memory_info)
    print_cpu_info(args, cpu_info, cpu_to_s)


if __name__ == '__main__':
    main()
