"""Functions to generate cgconfig files."""

import copy


from sp.cgroups import hierarchy
from sp.util.file import parsers


def indent(block, spaces=4):
    """Ident a block of text with number of spaces(default is 4)."""
    def indent_line(line):
        """Ident a single line with the number of spaces."""
        return ' ' * spaces + line
    lines = block.split('\n')
    lines = map(indent_line, lines)
    return '\n'.join(lines)


PRIORITIES = {
    'memory': {
        'use_hierarchy': 1,
        'swappiness': 2,
        'move_charge_at_immigrate': 3,
        'limit_in_bytes': 4,
        'memsw.limit_in_bytes': 5
    },
    'cpuset': {
        'cpus': 1,
        'mems': 2,
        'cpu_exclusive': 3
    }
}


def _priority_function_for(controller):

    def _priority_function(kvpair):
        option, _value = kvpair
        return PRIORITIES[controller].get(option, 10000)

    return _priority_function if controller in PRIORITIES else lambda _: 0


OPTIONS_GROUP_TEMPLATE = """{name} {{
{content}
}}"""


SLICE_TEMPLATE = """group {name} {{
{content}
}}"""


DEFAULT_SLICE_OPTIONS = {
    'cpuset': {
        'cpus': '',
        'mems': ''
    },
    'memory': {
        'use_hierarchy': 1,
        'move_charge_at_immigrate': 1,
        'limit_in_bytes': '',
    }
}


def gen_optionsgroups(optionsgroups):
    """Generates options groups configuration string from a dictionary."""
    def gen_optionsgroup(name_options_pair):
        """Generates options group configuration string from a dictionary."""
        name, options = name_options_pair
        prf = _priority_function_for(name)
        content = '\n'.join(
            "{0}.{1}=\"{2}\";".format(name, option, value)
            for option, value in sorted(options.items(), key=prf))
        return OPTIONS_GROUP_TEMPLATE.format(
            name=name,
            content=indent(content)
        )

    return '\n'.join(map(gen_optionsgroup, optionsgroups.items()))


def gen_slice(name, optionsgroups):
    """
    Generates slice configuration string from and options groups dictionary.
    """
    content = indent(gen_optionsgroups(optionsgroups))
    return SLICE_TEMPLATE.format(name=name, content=content)


def gen_default_slice(name, cpus, limit, numas, memsw):
    """Generates a slice string with default options and
    the given name, memory limit, cpus and numa nodes."""
    optionsgroups = copy.deepcopy(DEFAULT_SLICE_OPTIONS)
    optionsgroups['cpuset']['cpus'] = parsers.list_to_sranges(cpus)
    optionsgroups['cpuset']['mems'] = parsers.list_to_sranges(numas)
    optionsgroups['memory']['limit_in_bytes'] = limit
    if memsw:
        optionsgroups['memory']['memsw.limit_in_bytes'] = limit
    return gen_slice(name, optionsgroups)


def gen_sp_service_cpusubgroup(service, cpus, mems):
    """Generates storpool.slice/<service> cpuset subgroup string."""
    optionsgroups = {
        'cpuset': {
            'cpus': parsers.list_to_sranges(cpus),
            'mems': parsers.list_to_sranges(mems)
        }
    }
    return gen_slice('storpool.slice/' + service, optionsgroups)


def gen_sp_memory_subgroup(name, limit, memsw):
    """Generates storpool.slice memory subgroup string."""
    optionsgroups = {
        'memory': {
            'move_charge_at_immigrate': 3,
            'limit_in_bytes': limit,
        }
    }

    if memsw:
        optionsgroups['memory']['memsw.limit_in_bytes'] = limit

    return gen_slice('storpool.slice/' + name, optionsgroups)


def gen_sp_main(limit, cpus, numas, memsw):
    """Generates storpool.slice main options string."""
    optionsgroups = {
        'cpuset': {
            'cpus': parsers.list_to_sranges(cpus),
            'mems': parsers.list_to_sranges(numas),
            'cpu_exclusive': 1
        },
        'memory': {
            'swappiness': 0,
            'move_charge_at_immigrate': 3,
            'limit_in_bytes': limit,
            'use_hierarchy': 1
        }
    }

    if memsw:
        optionsgroups['memory']['memsw.limit_in_bytes'] = limit

    return gen_slice('storpool.slice', optionsgroups)


def gen_sp_slice(service_to_cpu, cpulist, limits, numas, memsw):
    """Generates storpool.slice string by adding services and memory strings
    to the main options string.
    """
    sp_all = [gen_sp_main(limits['all'], cpulist, numas, memsw)]

    for submem in ('common', 'alloc'):
        sp_all.append(gen_sp_memory_subgroup(submem, limits[submem], memsw))

    for service, cpus in service_to_cpu.items():
        sp_all.append(gen_sp_service_cpusubgroup(service, cpus, numas))

    return '\n\n'.join(sp_all)


def gen_slices_strs(slices_info, service_to_cpu, memsw=True):
    """Generate configuration strings for all slices.
    Returns a slicename-sliceconfig dictionary.
    """
    all_numas = hierarchy.get().cpuset_info()['.']['cpuset.mems']
    slice_strs = {}

    for slc, info in slices_info.items():
        if slc == 'storpool.slice':
            slice_strs[slc] = gen_sp_slice(service_to_cpu, info['cpuset'],
                                           info['limit'], all_numas, memsw)
        else:
            slice_strs[slc] = gen_default_slice(slc, info['cpuset'],
                                                info['limit'], all_numas,
                                                memsw)

    return slice_strs
