"""Defines holder classes for different cgtool configuration aspects."""

import os

from sp.disk import disklist as spdisk
from sp.netiface import ifaceutil, spiface
from sp import cpuinfo
from sp import installed
from sp import messages as m
from sp.util import limits as limutils
from sp.util import misc as utils

from ..util import servers
from ..util.mtype import get_mtype


def _installed_servers_count():
    installed_services = installed.services()
    for i in range(installed.MAX_SERVERS, 1, -1):
        if 'storpool_server_{n}'.format(n=i - 1) in installed_services:
            return i
    return bool('storpool_server' in installed_services)


MAX_SERVERS = max(7, _installed_servers_count())


class OptionsConfig(object):
    """Holder for script specific command line options."""
    # pylint: disable=too-few-public-methods

    def __init__(self):
        self.noop = None
        self.migrate = None
        self.interactive = None
        self.expand = None
        self.preserve_cpus = None
        self.preserve_limits = None

    def validate(self):
        """Validate the configuration."""
        assert isinstance(self.noop, bool)
        assert isinstance(self.migrate, bool)
        assert isinstance(self.interactive, bool)
        assert isinstance(self.expand, bool)
        assert isinstance(self.preserve_cpus, bool)
        assert isinstance(self.preserve_limits, bool)


class DirectoryConfig(object):
    """Holder for directory configuration."""

    def __init__(self):
        self.confdir = '/etc/cgconfig.d'
        self.cachedir = '/etc/storpool.conf.d'

    def __str__(self):
        return (
            'CONFDIR={0}\nCACHEDIR={1}\n'.format(self.confdir, self.cachedir)
        )

    def validate(self):
        """Validate the configuration."""
        assert os.path.isdir(self.confdir)
        assert os.path.isdir(self.cachedir)


class NetworkConfig(object):
    """Holder for network configuration."""

    def __init__(self):
        self.iface = None
        self.iface_acc = None

    def __str__(self):
        return (
            'IFACE={0}\nIFACE_ACC={1}\n'
            .format(self.iface, int(self.iface_acc))
        )

    def validate(self):
        """Validate the configuration."""
        assert ifaceutil.is_device(self.iface)
        assert isinstance(self.iface_acc, bool)


class MachineTypeConfig(object):
    """Holder for machine type configuration."""

    def __init__(self):
        self.converged = 0

    def __str__(self):
        return 'CONVERGED={0}'.format(int(self.converged))

    def validate(self):
        """Validate the configuration."""
        assert isinstance(self.converged, bool)


class ServicesConfig(object):
    """Holder for desired services configuration."""

    def __init__(self):
        self.servers = 0
        self.block = True
        self.iscsi = False
        self.mgmt = False
        self.bridge = False

    def __str__(self):
        def _attr_line(attr):
            return '{0}={1}\n'.format(attr.upper(), int(getattr(self, attr)))

        return ''.join(_attr_line(attr) for attr in
                       ('block', 'iscsi', 'mgmt', 'bridge', 'servers'))

    @property
    def list_services(self):
        """Construct a list of all services."""
        servs = ['rdma', 'beacon']
        servs.extend(servers.lst(self.servers))
        for service in ('block', 'iscsi', 'mgmt', 'bridge'):
            if getattr(self, service):
                servs.append(service)
        return servs

    def validate(self):
        """Validate the configuration."""
        assert isinstance(self.servers, int)
        assert isinstance(self.block, bool)
        assert isinstance(self.iscsi, bool)
        assert isinstance(self.mgmt, bool)
        assert isinstance(self.bridge, bool)


class CPUConfig(object):
    """Holder for cpu configuration."""

    def __init__(self):
        self.cores = None
        self.extra_thread = None
        self.numa_overflow = False

    @property
    def cores_tuple(self):
        """Cores, extra thread flag pair."""
        return (self.cores, self.extra_thread)

    @property
    def cores_str(self):
        """Cores config string as it would go into a config file."""
        return '{0}.5'.format(self.cores) if self.extra_thread \
            else str(self.cores)

    def __str__(self):
        return 'CORES={0}\nNUMA_OVERFLOW={1}\n' \
               .format(self.cores_str, int(self.numa_overflow))

    def validate(self):
        """Validate the configuration."""
        assert isinstance(self.cores, int)
        assert isinstance(self.extra_thread, bool)
        assert isinstance(self.numa_overflow, bool)
        assert self.cores > 0 or (self.cores == 0 and self.extra_thread)


class MemoryConfig(object):
    """Holder for memory configuration."""
    # pylint: disable=too-many-instance-attributes

    def __init__(self):
        self.set_cache_size = None
        self.sp_common_limit = None
        self.sp_alloc_limit = None
        self.system_limit = '2G'
        self.kernel_mem = '1G'
        self.user_limit = None
        self.machine_limit = None
        self.set_memsw = True

    def __str__(self):
        cfgs = 'SET_CACHE_SIZE={0}\n'.format(int(self.set_cache_size))
        cfgs += 'SET_MEMSW={0}\n'.format(int(self.set_memsw))
        cfgs += 'SP_COMMON_LIMIT={0}\n'.format(self.sp_common_limit)
        cfgs += 'SP_ALLOC_LIMIT={0}\n'.format(self.sp_alloc_limit)
        cfgs += 'SYSTEM_LIMIT={0}\n'.format(self.system_limit)
        cfgs += 'KERNEL_MEM={0}\n'.format(self.kernel_mem)
        if self.user_limit is not None:
            cfgs += 'USER_LIMIT={0}\n'.format(self.user_limit)
        if self.machine_limit is not None:
            cfgs += 'MACHINE_LIMIT={0}\n'.format(self.machine_limit)
        return cfgs

    def get_limits(self):
        """
        Constructs a slice-(subslices)-limit dictionary from the configuration.
        """
        limits = {
            'system': self.system_limit,
            'storpool': {
                'common': self.sp_common_limit,
                'alloc': self.sp_alloc_limit,
                'all': limutils.sum_limit_strs(self.sp_common_limit,
                                               self.sp_alloc_limit)
            }
        }
        if self.machine_limit is not None:
            limits['machine'] = self.machine_limit
        if self.user_limit is not None:
            limits['user'] = self.user_limit
        return limits

    def get_memory(self):
        """Constructs slice-limits dictionary only for the present limits."""
        mem = {
            'system': self.system_limit,
            'kernel': self.kernel_mem,
        }
        if self.sp_common_limit is not None:
            mem['storpool_common'] = self.sp_common_limit
        if self.sp_alloc_limit is not None:
            mem['storpool_alloc'] = self.sp_alloc_limit
        if self.machine_limit is not None:
            mem['machine'] = self.machine_limit
        if self.user_limit is not None:
            mem['user'] = self.user_limit
        return mem

    def validate(self):
        """Validate the configuration."""
        assert utils.is_string(self.sp_common_limit)
        assert utils.is_string(self.sp_alloc_limit)
        assert utils.is_string(self.system_limit)
        assert utils.is_string(self.kernel_mem)

        assert limutils.valid_limit_str(self.sp_common_limit)
        assert limutils.valid_limit_str(self.sp_alloc_limit)
        assert limutils.valid_limit_str(self.system_limit)
        assert limutils.valid_limit_str(self.kernel_mem)
        if self.machine_limit is not None:
            assert utils.is_string(self.machine_limit)
            assert limutils.valid_limit_str(self.machine_limit)
        if self.user_limit is not None:
            assert utils.is_string(self.user_limit)
            assert limutils.valid_limit_str(self.user_limit)


CONFIG_TMPL = """########## START CONFIG ##########
[cgtool]
{typeconf}
{serviceconf}
{memconf}
{netconf}
{cpuconf}
{dirconf}
########### END CONFIG ###########
"""


class ControlGroupToolConfig(object):
    """Holder for the complete configuration of the cgtool."""

    def __init__(self):
        self.optconf = None
        self.dirconf = None
        self.netconf = None
        self.typeconf = None
        self.serviceconf = None
        self.memconf = None
        self.cpuconf = None

    def __str__(self):
        return CONFIG_TMPL.format(typeconf=self.typeconf,
                                  serviceconf=self.serviceconf,
                                  memconf=self.memconf, netconf=self.netconf,
                                  cpuconf=self.cpuconf, dirconf=self.dirconf)

    @property
    def mtype(self):
        """
        Return a string representation of the machine type
        (client/server/combined).
        """
        return get_mtype(self.serviceconf, self.typeconf)

    def validate(self):
        """Validate the configuration."""
        self.optconf.validate()
        self.dirconf.validate()
        self.netconf.validate()
        self.typeconf.validate()
        self.serviceconf.validate()
        self.memconf.validate()
        self.cpuconf.validate()

    def sanity_check(self):
        """Perform a sanity check of the configuration."""
        self.validate()
        if self.serviceconf.servers not in range(MAX_SERVERS + 1):
            m.exit_msg('Invalid servers count')
        if self.serviceconf.servers == 0 and self.typeconf.converged:
            m.exit_msg('Converged setup with no servers')
        if not self.serviceconf.block:
            m.warn_msg('No block service')
        if not cpuinfo.get().hyperthreading() and self.cpuconf.extra_thread:
            m.exit_msg('Allocated single thread from a core, but there is no '
                       'hyperthreading on the machine. You may want to '
                       'allocate {n} cores.'
                       .format(n=self.cpuconf.cores * 2 + 1))
        if self.memconf.set_cache_size:
            before_msg = 'Cache size cannot be set.'
            after_msg = 'Please unset set_cache_size flag or fix the ' \
                        'servers configuration on the machine.'
            force_servers_count_match(self.serviceconf, before_msg, after_msg)

        limits_list = self.memconf.get_memory().values()
        all_mem = limutils.limit_str_to_mb(
            limutils.sum_limit_strs(*limits_list)
        )
        total_mem = utils.get_memtotal_kb() // 1024
        if total_mem < all_mem:
            m.exit_msg('Sum of memory limits ({0}) exceeds total machine '
                       'memory ({1})'.format(all_mem, total_mem))

        if self.netconf.iface_acc and not spiface.sp_has_acc():
            m.warn_msg('Current version of StorPool does not support interface'
                       ' acceleration.')
        if self.netconf.iface_acc and not spiface.has_acc(self.netconf.iface):
            m.warn_msg('Interface {0} does not support acceleration.'
                       .format(self.netconf.iface))


def force_servers_count_match(
        servconf,
        before_option_err_msg='',
        after_option_err_msg=''
):
    """
    Check if configuration server count matches the running server instances
    count and exit if not.
    """
    if before_option_err_msg:
        before_option_err_msg = '{0} '.format(before_option_err_msg)
    if after_option_err_msg:
        after_option_err_msg = ' {0}'.format(after_option_err_msg)
    cur_servers = spdisk.server_instances()
    cur_servers_str = str(cur_servers) if cur_servers else 'no'
    if servconf.servers != cur_servers:
        m.exit_msg('{0}SERVERS option in the config({1}) mismatches with '
                   'number of configured servers in storpool_initdisk({2}).{3}'
                   .format(before_option_err_msg, servconf.servers,
                           cur_servers_str, after_option_err_msg))
