"""Functions to generate cgtool configuration."""

from __future__ import division

import re

from sp import cpuinfo
from sp import installed
from sp import spmemory as hp
from sp import service as svc
from sp.cgroups import hierarchy as cgh
from sp.disk import disklist as spdisk
from sp.netiface import spiface
from sp import messages as m
from sp.util import limits as limutils
from sp.util import misc as utils

from .. import spcpu, spmem
from . import config

from .config import force_servers_count_match


NOT_FOUND_CFG_TMPL = "{option} was not found in config{action}"


def not_found_msg(option, action):
    """Construct a not-found-in-config message for the given option."""
    m.msg(NOT_FOUND_CFG_TMPL.format(option=option.upper(), action=action))


def get_services_conf(parser):
    """Construct ServicesConfig from the parser."""
    bins = svc.strip_prefix(installed.services())
    servconf = config.ServicesConfig()
    for service in ('block', 'iscsi', 'mgmt', 'bridge'):
        if parser.has_option('cgtool', service):
            setattr(servconf, service, parser.getboolean('cgtool', service))
        else:
            present = '' if service in bins else 'not '
            action = "; {b} is {p}installed, so will use {s}={v}" \
                     .format(b='storpool_{0}'.format(service),
                             s=service.upper(), p=present, v=service in bins)
            not_found_msg(service, action)
            setattr(servconf, service, service in bins)

    if parser.has_option('cgtool', 'servers'):
        servconf.servers = parser.getint('cgtool', 'servers')
    elif 'server' not in bins:
        servconf.servers = 0
        action = '; storpool_server is not installed, so will use SERVERS=0'
        not_found_msg('servers', action)
    else:
        inst = spdisk.server_instances()
        if not inst:
            action = '; storpool_server is installed, but the disks for ' \
                     'storpool are not initiated. Please specify the number ' \
                     'of server instances in the config file or initiate the' \
                     ' disks. Note that only specifying the number of ' \
                     'instances might not be sufficient.'
            m.exit_msg(NOT_FOUND_CFG_TMPL.format(
                option='SERVERS',
                action=action
            ))
        else:
            action = '; storpool_server is installed and {n} servers are ' \
                     'initiated, so will use SERVERS={n}'.format(n=inst)
            not_found_msg('servers', action)
            servconf.servers = inst
    return servconf


def get_attr(parser, attr, default_generator, getter='get'):
    """Get attribute from the 'cgtool' section of the parser.
    If not present, a default value will be generated with default_generator.
    A parser getter function may be specified (default is 'get', alternatives
    are 'getboolean', 'getfloat', 'getinteger').
    """
    if parser.has_option('cgtool', attr):
        getterfunc = getattr(parser, getter)
        attr_value = getterfunc('cgtool', attr)
    else:
        attr_value = default_generator()
        default_info = "; Will use {o}={v}" \
                       .format(o=attr.upper(), v=attr_value)
        not_found_msg(attr, default_info)
    return attr_value


def get_directory_conf(parser):
    """Construct DirectoryConfig from the parser."""
    dirconf = config.DirectoryConfig()
    dirconf.confdir = get_attr(parser, 'confdir', lambda: dirconf.confdir)
    dirconf.cachedir = get_attr(parser, 'cachedir', lambda: dirconf.cachedir)
    return dirconf


def get_network_conf(parser):
    """Construct NetworkConfig from the parser."""
    netconf = config.NetworkConfig()
    if not parser.has_option('cgtool', 'iface'):
        m.msg('Detecting network interface')
    netconf.iface = get_attr(parser, 'iface', spiface.first_sp_iface)
    if not parser.has_option('cgtool', 'iface_acc'):
        m.msg('Detecting if hardware acceleration is possible on {0}'
              .format(netconf.iface))
    netconf.iface_acc = get_attr(
        parser,
        'iface_acc',
        lambda: spiface.has_acc(netconf.iface),
        'getboolean'
    )
    return netconf


def _get_machine_type_conf(parser, servconf, optconf):
    mtconf = config.MachineTypeConfig()
    mtconf.converged = False
    if parser.has_option('cgtool', 'converged'):
        mtconf.converged = parser.getboolean('cgtool', 'converged')
    else:
        if servconf.servers > 0 and \
                parser.has_option('cgtool', 'machine_limit'):
            mtconf.converged = True
        if optconf.migrate and servconf.servers > 0 and \
                'machine.slice' in cgh.get().memory_info():
            mtconf.converged = True
        action = '; will use CONVERGED={0}'.format(mtconf.converged)
        not_found_msg('converged', action)
    return mtconf


def get_memory_and_type_confs(parser, optconf, servconf):
    """Construct MemoryConfig and MachineTypeConfig from the parser."""
    mtconf = _get_machine_type_conf(parser, servconf, optconf)
    memconf = config.MemoryConfig()

    if parser.has_option('cgtool', 'set_cache_size') and servconf.servers == 0:
        m.msg('SET_CACHE_SIZE option will be ignored for a client machine.')
        memconf.set_cache_size = False
    else:
        memconf.set_cache_size = get_attr(
            parser,
            'set_cache_size',
            lambda: not optconf.migrate,
            'getboolean'
        )

    if not parser.has_option('cgtool', 'system_limit'):
        bump = spmem.system_limit_bump()
        memconf.system_limit = limutils.sum_limit_strs(
            memconf.system_limit, bump
        )
        m.msg('Default system.slice limit is {0}; bump is {1}'
              .format(memconf.system_limit, bump))
    memconf.system_limit = get_attr(
        parser, 'system_limit', lambda: memconf.system_limit
    )

    if not parser.has_option('cgtool', 'kernel_mem'):
        hpmb = hp.all_hugepages_memory_mb()
        memconf.kernel_mem = limutils.sum_limit_strs(
            memconf.kernel_mem, '{0}M'.format(hpmb)
        )
    memconf.kernel_mem = get_attr(
        parser, 'kernel_mem', lambda: memconf.kernel_mem
    )

    if utils.is_systemd():
        m.msg('The machine runs systemd and needs a user slice')
        memconf.user_limit = get_attr(parser, 'user_limit', lambda: '2G')
    else:
        if parser.has_option('cgtool', 'user_limit'):
            m.warn_msg('Option USER_LIMIT will be ignored. The machine does '
                       'not run systemd')

    def _alloc_default_gen():
        return spmem.sp_alloc_limit(servconf.iscsi, servconf.mgmt,
                                    servconf.bridge)

    memconf.sp_alloc_limit = get_attr(
        parser, 'sp_alloc_limit', _alloc_default_gen
    )

    if servconf.servers == 0:  # client
        memconf.sp_common_limit = get_attr(
            parser,
            'sp_common_limit',
            lambda: spmem.sp_common_min_limit(memconf, servconf)
        )
        memconf.machine_limit = get_attr(
            parser, 'machine_limit', lambda: mem_left(memconf)
        )
    elif not mtconf.converged:  # server
        if parser.has_option('cgtool', 'machine_limit'):
            m.exit_msg('MACHINE_LIMIT option is set, but the machine is '
                       'dedicated server. Please check the configuration for '
                       'the tool')

        def _sp_common_def_gen():
            mleft = mem_left(memconf)
            min_default_limit = spmem.sp_common_min_limit(memconf, servconf)

            mleft_val = limutils.limit_str_to_mb(mleft)
            min_default_limit_val = limutils.limit_str_to_mb(min_default_limit)

            if mleft_val < min_default_limit_val:
                return min_default_limit
            return mleft

        memconf.sp_common_limit = get_attr(
            parser, 'sp_common_limit', _sp_common_def_gen
        )
    else:  # combined
        if parser.has_option('cgtool', 'machine_limit'):
            memconf.machine_limit = parser.get('cgtool', 'machine_limit')
            memconf.sp_common_limit = get_attr(
                parser, 'sp_common_limit', lambda: mem_left(memconf)
            )
        elif parser.has_option('cgtool', 'sp_common_limit'):
            memconf.sp_common_limit = parser.get('cgtool', 'sp_common_limit')
            memconf.machine_limit = get_attr(
                parser, 'machine_limit', lambda: mem_left(memconf)
            )
        else:
            force_servers_count_match(
                servconf, 'Cannot calculate storpool(common) memory limit.'
            )
            memconf.sp_common_limit = get_attr(
                parser,
                'sp_common_limit',
                lambda: spmem.sp_common_min_limit(memconf, servconf)
            )
            memconf.machine_limit = get_attr(
                parser, 'machine_limit', lambda: mem_left(memconf)
            )

    memconf.set_memsw = get_attr(
        parser, 'set_memsw', lambda: memconf.set_memsw, getter='getboolean'
    )

    return memconf, mtconf


def get_cpu_conf(parser, servconf, netconf, mtconf):
    """
    Construct CPUConfig from the parser and other already created
    configurations.
    """
    cpuconf = config.CPUConfig()
    cpuconf.numa_overflow = get_attr(
        parser,
        'numa_overflow',
        lambda: cpuconf.numa_overflow,
        'getboolean'
    )
    if parser.has_option('cgtool', 'cores'):
        cores = parse_cores(parser.get('cgtool', 'cores'))
        if cores is None:
            m.exit_msg('Invalid value for cores set in config.')
        cpuconf.cores = cores[0]
        cpuconf.extra_thread = cores[1]
    else:
        threads = spcpu.calculate_threads_needed(servconf, netconf, mtconf)
        has_hyper = cpuinfo.get().hyperthreading()
        cpuconf.cores = threads // 2 if has_hyper else threads
        cpuconf.extra_thread = bool(threads % 2) if has_hyper else False
        action = "; Will use CORES={c}".format(c=cpuconf.cores_str)
        not_found_msg('cores', action)
    return cpuconf


def validate_existing_limits(memconf):
    """Validate the limit strings of the memory config."""
    for lim_attr in ('sp_common_limit', 'sp_alloc_limit', 'kernel_mem',
                     'system_limit', 'user_limit', 'machine_limit'):
        limit = getattr(memconf, lim_attr)
        if limit is not None and not limutils.valid_limit_str(limit):
            m.exit_msg("{v} is not a valid value for {o}"
                       .format(v=limit, o=lim_attr.upper()))


def mem_left(memconf):
    """Calculate how much memory is left when all the limits in memconf are
    subtracted from the machine total memory. Returns a limit string.
    """
    validate_existing_limits(memconf)
    limits = memconf.get_memory().values()
    total_mb = utils.get_memtotal_kb() // 1024
    taken_mb = sum(limutils.limit_str_to_mb(limit) for limit in limits)
    if taken_mb >= total_mb:
        return '0M'
    return limutils.limit_to_gb_str((total_mb - taken_mb) / 1024)


def parse_cores(cores):
    """Parser for the value of CORES config parameter."""
    match = re.match(r"^(?P<cores>\d+)(?P<extra>\.5)?$", cores)
    if not match:
        return None

    cores, extra = int(match.group('cores')), match.group('extra') == '.5'
    if cores == 0 and not extra:
        return None

    return (cores, extra)
