"""Provides live-migration specific configuration validating checks."""

from __future__ import division


from sp.cgroups import filters as cgfilters
from sp.validations import cgroups as cgroups_validations
from sp.validations.check import Check as CGCheck
from sp.util import limits as limutils

from ..config import MAX_SERVERS
from ..util import spcg
from ..util import servers


def slice_new_limit_check(slice_info, slice_name, new_limit):
    """Returns a check that validates new memory limits are adequate."""
    errors = []
    warnings = []
    cur_limit_mb = slice_info['memory.limit_in_bytes'] // 1024 ** 2
    new_limit_mb = limutils.limit_str_to_mb(new_limit)

    if cgfilters.is_sp_subgroup(slice_name) and not slice_info['cgroup.procs']:
        return [], ['{0} has no running processes so it will be recreated to '
                    'avoid old memory accounted there'.format(slice_name)]

    if cur_limit_mb <= new_limit_mb:
        return [], []

    max_usage_mb = slice_info['memory.max_usage_in_bytes'] // 1024 ** 2
    cur_usage_mb = slice_info['memory.usage_in_bytes'] // 1024 ** 2
    if cur_usage_mb > new_limit_mb:
        errors.append('memory:{0} new limit is less than current usage'
                      .format(slice_name))
    else:
        margin = cur_usage_mb / 5  # 20%
        if cur_usage_mb + margin > new_limit_mb:
            errors.append('if memory:{0} is set to new limit, it will have '
                          'more than 80% usage which is dangerous'
                          .format(slice_name))
        if max_usage_mb > new_limit_mb:
            warnings.append('memory:{0} new limit is less than max usage'
                            .format(slice_name))
    return errors, warnings


def obsolete_services_check(cfg, cpuset_info):
    """
    Returns a check that looks for services that are no longer needed
    (in the configuration), but are still actively used.
    """
    def has_procs(service):
        """
        Indicates whether the storpool subslice for the service has running
        processes.
        """
        srv_slc = 'storpool.slice/{0}'.format(service)
        return bool(cpuset_info[srv_slc]['cgroup.procs'])

    err_tmp = 'cpuset:storpool.slice/{0} is not needed in the new setup, but '\
              'there are processes acounted there'
    active = filter(has_procs, spcg.current_services(cpuset_info))
    unneeded_servers = servers.lst(MAX_SERVERS)[cfg.serviceconf.servers:]
    unneeded_other = [s for s in ('mgmt', 'bridge', 'iscsi', 'block')
                      if getattr(cfg.serviceconf, s) == 0]
    unneeded_active = set(unneeded_servers + unneeded_other) & set(active)
    return [err_tmp.format(service) for service in unneeded_active], []


def get_migration_check(cfg, cpuset_info, memory_info, _new_sp_cpus):
    """Returns a whole 'ok to migrate' check."""
    base_chk, _cpu_chk, _mem_chk = cgroups_validations.get_matching_check(
        cpuset_info,
        memory_info
    )
    checks = {
        'base': base_chk,
        'obsolete_services': CGCheck([base_chk], obsolete_services_check, cfg,
                                     cpuset_info),
    }

    def _add_slice_new_limit_check(slice_name, new_limit):
        checks[slice_name] = CGCheck([base_chk], slice_new_limit_check,
                                     memory_info[slice_name], slice_name,
                                     new_limit)

    new_limits = cfg.memconf.get_limits()
    for slc, slc_limit in new_limits.items():
        slc_name = slc + '.slice'
        if slc == 'storpool':
            _add_slice_new_limit_check(slc_name, slc_limit['all'])
            for subslc in ('common', 'alloc'):
                full_name = slc_name + '/' + subslc
                if full_name in memory_info:
                    _add_slice_new_limit_check(full_name, slc_limit[subslc])
        elif slc_name in memory_info:
            _add_slice_new_limit_check(slc_name, slc_limit)
    return CGCheck(checks.values(), lambda: ([], []))
