"""Defines migrating functions for the live migration of the configuration.
The live migration process is as follows:
    1. Perform pre-migration configuration checks. If the checks fail,
    the migration is unsucessful and is aborted. (If ran in interactive mode
    check results can be ignored).

    2. Migrate the cpuset cgroups:
        2.1 Try to expand storpool.slice if needed so that it has equal or more
            cpus that the desired configuration
        2.2 Clean the storpool.slice from unneeded subslices.
        2.3 Transfer cpus so that storpool.slice has the desired cpuset. This
        step has three separate scenarios:
            I) The cpuset is already the desired one - do nothing :)
            II) The current and new cpusets have enough common cpus so that
            storpool services can be shrunk to use only them while exchanging
            the others. This is the so-called 'normal' cpuset migration. It
            proceeds as follows:
               II.1) StorPool is shrunk to use only the common cpus
               II.2) Freed cpus are removed from the storpool.slice
               II.3) Freed cpus are exchanged recursively with all other
               slices, that possess cpus required for StorPool. First, the
               freed cpus are added to the other slice cpuset and then the cpus
               needed for storpool are taken from that cpuset. It should be
               noted that the amount of given and taken cpus is exactly the
               same.
               II.4) storpool.slice is given the cpus removed from of the other
               slices so that it has the desired cpuset
            III) The current and new cpusets do not have enough common cpus.
            This is the most complex situation. In this case storpool.slice is
            first migrated to an intermediate cpuset that has enough common
            cpus with both the current and the new one. Then the storpool.slice
            is migrated to the new cpuset. These two migrations are done as
            described in II), because they both meet the requirements.
        2.4 Rebalance the storpool.slice so that each service goes to the
        desired cpu and StorPool is no longer shrunk. Also, create
        child groups if necessary and classify services to the right ones.
        2.5 Clean the old/obsolete mgmt.slice cpuset if it exists.
    3. Migrate the memory cgroups:
        3.1 For each of the memory slices, create it if it does not exist and
        set its limit to the desired one.
        3.2 If storpool.slice/{common,alloc} are newly created, classify the
        corresponding storpool processes to each of them.
        3.3 Remove the obsoleted mgmt.slice if it exists.
        3.4 Remove the machine.slice if the machine is migrated from client or
        hyperconverged setup to dedicated server.

    * Try to roll back the changes already made if any of the migration
    steps fails.
"""
from __future__ import division


import collections
import itertools
import math


from sp.cgroups import filters
from sp.process import sppidutil
from sp import messages as m
from sp import cpuinfo
from sp.util import limits as limutils
from sp.util.misc import query_yes_no, list2d_flatten
from sp.util.file import parsers

from .transaction import Transaction, TransactionError
from . import migrationcheck
from ..util import spcg


def sp_proc_info(ctrl_info):
    """Returns a service-pids dictionary for the storpool services accounted
    in the given controller info.
    """
    info = collections.defaultdict(list)
    for grp in ctrl_info:
        if filters.is_sp_cgroup(grp):
            pids = ctrl_info[grp]['cgroup.procs']
            for pid, status in sppidutil.recognise(pids).items():
                info[status].append(pid)
    return info


class MigrationError(Exception):
    """Base error for migration problems."""


def cpu_to_slices(cpuset_info):
    """Creates cpu-slices dictionary from the cpuset info."""
    cpu_slices = collections.defaultdict(list)

    for grp in filter(filters.is_not_sp_direct_root_child, cpuset_info):
        for cpu in cpuset_info[grp]['cpuset.cpus']:
            cpu_slices[cpu].append(grp)

    return cpu_slices


def expand_storpool_cpus(transaction, new_sp_cpus):
    """Exapand the StorPool cpuset to the desired one if possible."""
    cpuset_info = transaction.hierarchy.cpuset_info()
    current_sp_cpus = cpuset_info['storpool.slice']['cpuset.cpus']
    needed = len(new_sp_cpus) - len(current_sp_cpus)
    if needed <= 0:
        return True
    prefered = sorted(set(new_sp_cpus) - set(current_sp_cpus))[:needed]

    for grp in filter(filters.is_not_sp_direct_root_child, cpuset_info):
        left = set(cpuset_info[grp]['cpuset.cpus']) - set(prefered)
        if not left:
            return False

    m.msg('Chose {0} to expand StorPool with.'.format(prefered))
    for grp in filter(filters.is_not_sp_direct_root_child, cpuset_info):
        m.msg('Removing {0} from {1}...'.format(prefered, grp))
        transaction.remove_cpus(grp, prefered)
    m.msg('Adding {0} to SP slice...'.format(prefered))
    transaction.add_cpus_nosubgroups('storpool.slice', prefered)
    return True


def migrate_cpus(transaction, cfg, new_sp_cpus, s_to_cpu):
    """
    Applies the given cpuset configuration to the machine via the given
    transaction.
    """
    # pylint: disable=too-many-locals
    # pylint: disable=too-many-statements
    cpuset_info = transaction.hierarchy.cpuset_info()

    if len(new_sp_cpus) > len(cpuset_info['storpool.slice']['cpuset.cpus']):
        m.msg('Storpool.slice needs expanding before migrating.')
        if not expand_storpool_cpus(transaction, new_sp_cpus):
            m.exit_msg('Could not expand StorPool slice.')

    _sp_cpus = cpuset_info['storpool.slice']['cpuset.cpus']
    if (
            len(new_sp_cpus) == 1
            and len(_sp_cpus) == 1
            and _sp_cpus[0] != new_sp_cpus[0]
    ):
        m.exit_msg(
            'Cannot migrate StorPool from single cpu to a different single '
            'cpu. (You can temporarily migrate to 2+ cpus first.)'
        )

    cpuset_info = transaction.hierarchy.cpuset_info()
    current_sp_cpus = cpuset_info['storpool.slice']['cpuset.cpus']
    to_leave_sp = sorted(set(current_sp_cpus) - set(new_sp_cpus))
    to_come_sp = sorted(set(new_sp_cpus) - set(current_sp_cpus))
    to_stay_sp = sorted(set(current_sp_cpus) & set(new_sp_cpus))

    current_cpu_to_s = spcg.current_cpu_to_service(cpuset_info)
    current_services = list(set(list2d_flatten(current_cpu_to_s.values())))

    m.msg('Cleaning up cpuset:storpool.slice/ not needed subgroups...')
    for grp in filter(filters.is_sp_subgroup, cpuset_info):
        service = grp.split('/')[1]
        if service not in s_to_cpu.keys():
            transaction.remove_group('cpuset', grp)
            current_services.remove(service)

    def cpus_needed_to_shrink(services):
        """Calculates the minimum number of cpus, to which storpool services
        can be shrunk.
        """
        return int(math.ceil(len(services) / 3))

    # helper function for usual cases
    def normal_migration(cur_services, to_stay_sp, to_leave_sp, to_come_sp):
        """Performs cpus transfers in and out of storpool's cpuset to reach
        the desired cpuset by the new configuration. Requires the current and
        the new cpusets to have enough cpus in common so services can be
        shrunk to use only them while transferring.
        """
        def _give_freed_to_machine_only():
            migrating_to_server = (cfg.mtype == 'server')
            exists = transaction.hierarchy.group_exists(
                'cpuset', 'machine.slice'
            )
            has_separate_cpus = False
            if exists:
                cpuset = set(cpuset_info['machine.slice']['cpuset.cpus'])
                sys_cpuset = set(cpuset_info['system.slice']['cpuset.cpus'])
                has_separate_cpus = cpuset != sys_cpuset

            return exists and not migrating_to_server and has_separate_cpus

        def _generate_slices_swaps():
            cpu_slices = cpu_to_slices(cpuset_info)
            slice_swap = collections.defaultdict(
                lambda: {'take': [], 'give': []}
            )

            given_cpus = set()
            for cpu_to_take in to_come_sp:
                cpu_to_give = to_leave_sp.pop(0)

                for cg_slice in cpu_slices[cpu_to_take]:
                    slice_swap[cg_slice]['take'].append(cpu_to_take)
                    slice_swap[cg_slice]['give'].append(cpu_to_give)

                given_cpus.add(cpu_to_give)

            rem_to_be_given = set(to_leave_sp) - given_cpus

            if _give_freed_to_machine_only():
                slice_swap['machine.slice']['give'].extend(rem_to_be_given)
            else:
                for cg_slice in slice_swap:
                    slice_swap[cg_slice]['give'].extend(rem_to_be_given)

            return slice_swap

        m.msg('Shrinking SP services to use only {0} cpus... '
              .format(to_stay_sp))
        for service, cpu in zip(cur_services, itertools.cycle(to_stay_sp)):
            transaction.set_option('cpuset', 'storpool.slice/' + service,
                                   'cpuset.cpus', str(cpu) + '\n')

        m.msg('Removing {0} cpus from SP slice...'.format(to_leave_sp))
        transaction.remove_cpus('storpool.slice', to_leave_sp)

        for cg_slice, swap in _generate_slices_swaps().items():
            m.msg('Giving {0} cpus to {1}...'.format(swap['give'], cg_slice))
            transaction.add_cpus(cg_slice, swap['give'])
            m.msg('Taking {0} cpus from {1}...'.format(swap['take'], cg_slice))
            transaction.remove_cpus(cg_slice, swap['take'])

        m.msg('Adding {0} cpus to SP slice...'.format(to_come_sp))
        transaction.add_cpus_nosubgroups('storpool.slice', to_come_sp)

    if to_stay_sp == current_sp_cpus:  # to_leave_sp is empty
        # to_come_sp is empty too!
        m.msg('Old and new cpusets are the same.')
    elif len(to_stay_sp) >= cpus_needed_to_shrink(current_services):
        m.msg('SP can be shrunk to use only part of its current cpus.')
        normal_migration(current_services, to_stay_sp, to_leave_sp, to_come_sp)
    else:
        m.msg('SP cannot be shrunk to use only part of its current cpus.'
              'Migration will be made in two steps.')
        transfered_leave_stay = []
        wont_come = []
        while len(to_stay_sp) < cpus_needed_to_shrink(current_services):
            cpu = to_leave_sp.pop()
            to_stay_sp.append(cpu)
            transfered_leave_stay.append(cpu)
            if len(to_come_sp) > len(transfered_leave_stay):
                wont_come.append(to_come_sp.pop())
        normal_migration(current_services, to_stay_sp, to_leave_sp, to_come_sp)

        to_stay_sp = sorted(
            (set(to_stay_sp) - set(transfered_leave_stay)) | set(to_come_sp)
        )
        to_leave_sp = transfered_leave_stay
        to_come_sp = wont_come
        normal_migration(current_services, to_stay_sp, to_leave_sp, to_come_sp)

    rebalance_sp_slice(transaction, s_to_cpu)

    if cfg.mtype != 'server' and \
            not transaction.hierarchy.group_exists('cpuset', 'machine.slice'):
        m.msg('Creating cpuset:machine.slice...')
        machine_cpus = [cpu for cpu in cpuinfo.get().list_all_cpus()
                        if cpu not in new_sp_cpus]
        transaction.create_cpuset_group('machine.slice', cpus=machine_cpus)

    # clean
    if transaction.hierarchy.group_exists('cpuset', 'mgmt.slice'):
        m.msg('Removing old and obsolete cpuset:mgmt.slice...')
        transaction.remove_group('cpuset', 'mgmt.slice')


def rebalance_sp_slice(transaction, s_to_cpu):
    """Provided that the storpool's cpuset is already the desired one, appplies
    the service-per-cpu configuration to the slice. Child cgroups will be
    created if needed and processes will be moved to them.
    """
    m.msg('Applying new cpu bindings to storpool.slice/* ...')
    cpuset_info = transaction.hierarchy.cpuset_info(with_procs=True)
    proc_info = sp_proc_info(cpuset_info)
    for service, cpus in s_to_cpu.items():
        group = 'storpool.slice/{0}'.format(service)
        if group in cpuset_info:
            transaction.set_option(
                'cpuset',
                group,
                'cpuset.cpus',
                parsers.list_to_sranges(cpus) + '\n'
            )
        else:
            m.msg('Creating cpuset:{0}...'.format(group))
            transaction.create_cpuset_group(group, cpus)
            for pid in proc_info.get(service, []):
                m.msg('Moving pid {0} to new cpuset:{1}'.format(pid, group))
                transaction.classify('cpuset', group, pid)


def create_memory_group(transaction, name):
    """
    Creates a memory cgroup with the given swappiness value via the given
    transaction.
    """
    m.msg('Creating memory:{0}...'.format(name))
    transaction.create_group('memory', name)
    transaction.set_option('memory', name, 'memory.use_hierarchy', '1\n')
    transaction.set_option('memory', name, 'memory.move_charge_at_immigrate',
                           '3\n')


def set_memory_limit(transaction, slc_name, limit):
    """
    Applies the given memory limit to the given slice via the given
    transaction.
    """
    memory_info = transaction.hierarchy.memory_info(with_procs=True)

    if 'memory.memsw.limit_in_bytes' not in memory_info[slc_name]:
        m.msg('Setting memory:{0} new limit...'.format(slc_name))
        transaction.set_option('memory', slc_name, 'memory.limit_in_bytes',
                               limit + '\n')
        return

    increasing = False
    decreasing = True
    if slc_name in memory_info:
        cur_lim_b = memory_info[slc_name]['memory.limit_in_bytes']
        cur_lim_mb = cur_lim_b // 1024 ** 2
        new_lim_mb = limutils.limit_str_to_mb(limit)
        increasing = cur_lim_mb < new_lim_mb
        decreasing = cur_lim_mb > new_lim_mb

    if increasing:
        m.msg('Setting memory:{0} new limit...'.format(slc_name))
        transaction.set_option('memory', slc_name,
                               'memory.memsw.limit_in_bytes', limit + '\n')
        transaction.set_option('memory', slc_name,
                               'memory.limit_in_bytes', limit + '\n')

    if decreasing:
        m.msg('Setting memory:{0} new limit...'.format(slc_name))
        transaction.set_option('memory', slc_name,
                               'memory.limit_in_bytes', limit + '\n')
        transaction.set_option('memory', slc_name,
                               'memory.memsw.limit_in_bytes', limit + '\n')


def migrate_memory(transaction, cfg):
    """
    Applies the given memory configuration to the machine via the given
    transaction.
    """
    memory_info = transaction.hierarchy.memory_info(with_procs=True)
    new_limits = cfg.memconf.get_limits()

    def _migrate_sp_slice(new_limit):
        m.msg('Migrating memory:storpool.slice...')
        set_memory_limit(transaction, 'storpool.slice', new_limit['all'])
        for subslc in ('common', 'alloc'):
            full_name = 'storpool.slice/{0}'.format(subslc)
            if full_name not in memory_info:
                create_memory_group(transaction, full_name)
            # the following is to avoid dumb cgroups not cleaning when empty
            if full_name in memory_info and \
                    not memory_info[full_name]['cgroup.procs']:
                transaction.remove_group('memory', full_name)
                create_memory_group(transaction, full_name)
            set_memory_limit(transaction, full_name, new_limit[subslc])

    for slc, new_limit in new_limits.items():
        if slc == 'storpool':
            _migrate_sp_slice(new_limit)
        else:
            slc_name = slc + '.slice'
            if slc_name not in memory_info:
                create_memory_group(transaction, slc_name)
            set_memory_limit(transaction, slc_name, new_limit)

    proc_info = sp_proc_info(memory_info)
    # move procs
    alloc_pids = []
    if 'storpool.slice/alloc' not in memory_info.keys():
        m.msg('Moving pids to new memory:storpool.slice/alloc...')
        for srv in set(['mgmt', 'iscsi', 'bridge']) & set(proc_info.keys()):
            for pid in proc_info[srv]:
                transaction.classify('memory', 'storpool.slice/alloc', pid)
                alloc_pids.append(pid)

    if 'storpool.slice/common' not in memory_info.keys():
        m.msg('Moving pids to new memory:storpool.slice/common...')
        for pid in memory_info['storpool.slice']['cgroup.procs']:
            if pid not in alloc_pids:
                transaction.classify('memory', 'storpool.slice/common', pid)
    # clean
    if transaction.hierarchy.group_exists('memory', 'mgmt.slice'):
        m.msg('Removing old memory:mgmt.slice...')
        transaction.remove_group('memory', 'mgmt.slice')
    if cfg.mtype == 'server' and 'machine.slice' in memory_info.keys():
        m.msg('Removing old machine.slice, because new setup is server...')
        transaction.remove_group('cpuset', 'machine.slice')
        transaction.remove_group('memory', 'machine.slice')


def apply_config(cfg, hierarchy, sp_cpus, s_to_cpu):
    """Tries to apply the given configuration to the machine using the given
    cgroups hierarchy(may be mocked) Return True on succes, False on failure.
    MigrationError will be thrown if the migrating transaction fails AND its
    rollback also fails. MigrationError indicates that the machine may be in
    'broken' state.
    """
    cpuset_info, memory_info = hierarchy.cpuset_memory_info()
    check = migrationcheck.get_migration_check(cfg, cpuset_info, memory_info,
                                               sp_cpus)
    check.run()
    map(m.warn_msg, check.get_all_warnings())
    check_errors = check.get_all_errors()
    stop = bool(check_errors) if not cfg.optconf.interactive else False
    for check_err in check_errors:
        m.err_msg(check_err)
        if cfg.optconf.interactive and \
                query_yes_no('Ignore that error?', default='no') == 'no':
            stop = True
            break

    if stop:
        m.warn_msg('Setup checks failed.')
        return False

    if len(sp_cpus) > len(cpuset_info['storpool.slice']['cpuset.cpus']) \
       and not cfg.optconf.expand:
        m.exit_msg('New configuration requires more cpus than currently '
                   'available for StorPool. If you really want to take cpus '
                   'from other slices for StorPool, please set the '
                   '-E(--expand) option.')

    transaction = Transaction(hierarchy)
    try:
        m.msg('Migrating the cpuset slices...')
        migrate_cpus(transaction, cfg, sp_cpus, s_to_cpu)
        m.msg('Migrating the memory slices...')
        migrate_memory(transaction, cfg)
        m.msg('Migration done.')
        return True
    except TransactionError as err:
        m.err_msg(err)
        m.warn_msg("==== MIGRATION FAILED ====")
        m.warn_msg("=== TRYING TO ROLLBACK ===")
        try:
            transaction.rollback()
            m.warn_msg("=== ROLLBACK SUCCESSFUL ===")
            return False
        except TransactionError as err:
            m.warn_msg("===== ROLLBACK FAILED =====")
            raise MigrationError(err)
