"""
Provides a transactional mechanism to make changes to the cgroup hierarchy.
"""

from sp import cpuinfo
from sp import messages as m
from sp.process import pidutil
from sp.cgroups import hierarchy as cgh
from sp.util.file.parsers import sranges_to_list, list_to_sranges


class Operation(object):
    """Single transaction action."""

    def __init__(self):
        self.done = False

    def _do(self):
        raise NotImplementedError()

    def _undo(self):
        raise NotImplementedError()

    def perform(self):
        """Execute the operation."""
        if not self.done:
            self._do()
        self.done = True

    def rollback(self):
        """Rollback/undo the operation."""
        if self.done:
            self._undo()
            self.done = False


class SetOptionOperation(Operation):
    """Operation to set a cgroup property.
    Old value is cached and restored on rollback.
    """

    # pylint: disable=too-many-arguments
    def __init__(self, hierarchy, controller, group, option, value):
        Operation.__init__(self)
        self.hierarchy = hierarchy
        self.controller = controller
        self.group = group
        self.option = option
        self.new_value = value
        self.old_value = None

    def _do(self):
        self.old_value = self.hierarchy.get_option(self.controller, self.group,
                                                   self.option)
        if self.old_value != self.new_value:
            self.hierarchy.set_option(self.controller, self.group, self.option,
                                      self.new_value)

    def _undo(self):
        if self.old_value != self.new_value:
            self.hierarchy.set_option(self.controller, self.group, self.option,
                                      self.old_value)


class CreateGroupOperation(Operation):
    """Operation to create a new cgroup. The group is remove on rollback."""

    def __init__(self, hierarchy, controller, group):
        Operation.__init__(self)
        self.hierarchy = hierarchy
        self.controller = controller
        self.group = group

    def _do(self):
        self.hierarchy.create_group(self.controller, self.group)

    def _undo(self):
        self.hierarchy.remove_group(self.controller, self.group)


class RemoveGroupOperation(Operation):
    """Operation to remove a cgroup. The group is restored on rollback.
    Group options specified in the CGroupsHierarchy.options[<controller>]['RW']
    list will be cached and restored on rollback.
    """

    CTRL_OPTS = {
        'cpuset': cgh.CGroupsHierarchy.options['cpuset']['RW'],
        'memory': cgh.CGroupsHierarchy.options['memory']['RW'],
    }

    def __init__(self, hierarchy, controller, group):
        assert controller in RemoveGroupOperation.CTRL_OPTS.keys()
        Operation.__init__(self)
        self.hierarchy = hierarchy
        self.controller = controller
        self.group = group
        self.options_cache = {}

    def _do(self):
        for opt in RemoveGroupOperation.CTRL_OPTS[self.controller]:
            try:
                self.options_cache[opt] = self.hierarchy.get_option(
                    self.controller,
                    self.group,
                    opt
                )
            except ValueError:
                pass
        self.hierarchy.remove_group(self.controller, self.group)

    def _undo(self):
        self.hierarchy.create_group(self.controller, self.group)
        for opt, val in self.options_cache.items():
            self.hierarchy.set_option(self.controller, self.group, opt, val)


class ClassifyOperation(Operation):
    """Operation to move a process to another cgroup.
    On rollback, the process and its newly spawned children (spawned after the
    move) will be classified back into the original group.
    """

    def __init__(self, hierarchy, controller, group, pid):
        Operation.__init__(self)
        self.hierarchy = hierarchy
        self.controller = controller
        self.new_group = group
        self.old_group = None
        self.children_before_move = []
        self.pid = pid
        self.moved = False

    def _children_in_cgroup(self, group):
        return [pid for pid in
                self.hierarchy.group_procs(self.controller, group)
                if pidutil.get_parent(pid) == self.pid]

    def _classify(self, new_grp, pid):
        try:
            self.hierarchy.safe_classify(self.controller, new_grp, pid)
            return True
        except IOError as err:
            m.warn_msg('Could not move pid {p} to {g}. Err: {e}'.format(
                p=pid, g=new_grp, e=err
            ))
            return False

    def _do(self):
        self.old_group = pidutil.cgroup(self.pid, self.controller)
        self.children_before_move = self._children_in_cgroup(self.old_group)
        self.moved = self._classify(self.new_group, self.pid)

    def _undo(self):
        if not self.moved:
            return
        self._classify(self.old_group, self.pid)
        # move new spawned succesors
        children = self._children_in_cgroup(self.new_group)
        new_children = set(children) - set(self.children_before_move)
        new_succesors = [pid for pid in
                         self.hierarchy.group_procs(self.controller,
                                                    self.new_group)
                         if set(pidutil.ancestors_of(pid)) & new_children]
        for pid in new_succesors + list(new_children):
            self._classify(self.old_group, pid)


class TransactionError(Exception):
    """Base error for transaction problems."""


class Transaction(object):
    """A list of transaction operations that can be rollbacked together."""

    def __init__(self, hierarchy):
        self.operations = []
        self.failed = False
        self.hierarchy = hierarchy

    def _add_and_perform(self, new_operations):
        if self.failed:
            raise TransactionError()
        try:
            for operation in new_operations:
                self.operations.append(operation)
                operation.perform()
        except (ValueError, OSError, IOError) as err:
            self.failed = True
            raise TransactionError(err)

    def rollback(self):
        """Rollback all the operations performed successfully so far."""
        try:
            for operation in reversed(self.operations):
                operation.rollback()
        except (ValueError, OSError, IOError) as err:
            raise TransactionError(err)

    def create_group(self, controller, group):
        """Execute an operation to create a cgroup."""
        operations = [CreateGroupOperation(self.hierarchy, controller, group)]
        self._add_and_perform(operations)

    def remove_group(self, controller, group):
        """Execute an operation to remove a cgroup."""
        subgrps = sorted(self.hierarchy.subgroups(controller, group),
                         reverse=True)
        operations = [RemoveGroupOperation(self.hierarchy, controller, subgrp)
                      for subgrp in subgrps]
        self._add_and_perform(operations)

    def set_option(self, controller, group, option, value):
        """Execute an operation to set a cgroup's option to the given value."""
        assert option != 'cgroup.procs'
        assert option != 'cgroup.tasks'
        operations = [SetOptionOperation(self.hierarchy, controller, group,
                                         option, value)]
        self._add_and_perform(operations)

    def classify(self, controller, group, pid):
        """
        Execute an operation to classify the given process to the given cgroup.
        """
        operations = [ClassifyOperation(self.hierarchy, controller, group,
                                        pid)]
        self._add_and_perform(operations)

    def create_cpuset_group(self, group, cpus=None, mems=None):
        """
        Execute a few operations to create a cpuset group with the desired
        properties.
        """
        if cpus is None:
            cpus = range(0, cpuinfo.get().cpu_count())
        if mems is None:
            mems = self.hierarchy.cpuset_info()['.']['cpuset.mems']

        cpus = list_to_sranges(cpus) + '\n'
        mems = list_to_sranges(mems) + '\n'
        operations = [
            CreateGroupOperation(self.hierarchy, 'cpuset', group),
            SetOptionOperation(
                self.hierarchy, 'cpuset', group, 'cpuset.cpus', cpus
            ),
            SetOptionOperation(
                self.hierarchy, 'cpuset', group, 'cpuset.mems', mems
            )
        ]
        self._add_and_perform(operations)

    def remove_cpus(self, group, to_remove_cpus):
        """Execute a few operations to remove the given cpus from the
        cgroup's cpuset. Cpus will be removed recursively from the cgroup's
        child groups before they are removed from the cgroup.
        """
        to_remove_cpus = sorted(to_remove_cpus)
        subgrps = sorted(self.hierarchy.subgroups('cpuset', group),
                         reverse=True)
        operations = [
            _remove_cpus_single_grp(self.hierarchy, subgrp, to_remove_cpus)
            for subgrp in subgrps
        ]
        self._add_and_perform(operations)

    def add_cpus(self, group, to_add_cpus):
        """Execute a few operations to add the given cpus from the
        cgroup's cpuset. Cpus will be added recursively to the cgroup's
        child groups after they are added to the cgroup.
        """
        to_add_cpus = sorted(to_add_cpus)
        subgrps = sorted(self.hierarchy.subgroups('cpuset', group))
        operations = [_add_cpus_single_grp(self.hierarchy, subgrp, to_add_cpus)
                      for subgrp in subgrps]
        self._add_and_perform(operations)

    def add_cpus_nosubgroups(self, group, to_add_cpus):
        """Execute and operation to add the given cpus to the cgroup's cpuset.
        Not recursive, so child groups will not be affected.
        """
        operations = [_add_cpus_single_grp(self.hierarchy, group, to_add_cpus)]
        self._add_and_perform(operations)


def _remove_cpus_single_grp(hierarchy, grp, to_remove_cpus):
    cpus = sranges_to_list(hierarchy.get_option('cpuset', grp, 'cpuset.cpus'))
    new_cpus = sorted(cpu for cpu in cpus if cpu not in to_remove_cpus)
    return SetOptionOperation(hierarchy, 'cpuset', grp, 'cpuset.cpus',
                              list_to_sranges(new_cpus) + '\n')


def _add_cpus_single_grp(hierarchy, grp, to_add_cpus):
    cpus = sranges_to_list(hierarchy.get_option('cpuset', grp, 'cpuset.cpus'))
    new_cpus = sorted(set(cpus + to_add_cpus))
    return SetOptionOperation(hierarchy, 'cpuset', grp, 'cpuset.cpus',
                              list_to_sranges(new_cpus) + '\n')
