"""The module can generate cpu-services configurations for the machine."""

from __future__ import division
import collections
import copy
import functools
import itertools


from storpool import spconfig


from sp import cpuinfo
from sp.netiface import ifaceutil
from sp import messages as m
from sp.util import decorators as spd
from sp.util import misc as utils

from .config import MAX_SERVERS
from .util import servers
from .util.mtype import get_mtype


@spd.static_vars(flag=None)
def uses_ksleep(reread=False):
    """Indicates wether StorPool uses ksleep or not."""
    if uses_ksleep.flag is None or reread:
        cfg = spconfig.SPConfig()
        method = cfg.get('SP_SLEEP_TYPE', '')
        uses_ksleep.flag = method in ('', 'ksleep')
    return uses_ksleep.flag


def block_is_cpu_friendly():
    """
    Indicates wether the storpool_block is cpu-friendly to its threadmates.
    """
    return uses_ksleep()


@spd.static_vars(flag=None)
def bridge_needs_two_cpus(reread=False):
    """
    Indicates wether the storpool_bridge needs two cpus.
    """
    if bridge_needs_two_cpus.flag is None or reread:
        cfg = spconfig.SPConfig()
        method = cfg.get('SP_BRIDGE_SLEEP_TYPE', '')
        bridge_needs_two_cpus.flag = (method and method != 'ksleep')
    return bridge_needs_two_cpus.flag


def get_sp_cpus(iface, cores, allow_nonlocal):
    """Returns a list of cpus to be used for StorPool.
    Cpus will be taken from the local cpulist of the provided interface.
    `cores` is a pair, containing the number of cores and a flag, which
    indicates whether to use an extra hyperthread.
    Note: Cpus on the first core of a numa node will be avoided.
    """
    # pylint: disable=too-many-locals
    def _not_enough_local_cores_msg(needed, available):
        status = 'allowed' if allow_nonlocal else 'not allowed'
        return 'There are not enough cores in the network interface local ' \
               'cpulist {n} > {a}. Using nonlocal cpus is {s} in the config.' \
               .format(n=needed, a=available, s=status)

    def _not_enough_cores_msg(needed, available):
        return 'There are not enough cores on the machine {n} > {a}' \
               .format(n=needed, a=available)

    cpui = cpuinfo.get()
    try:
        localcpus = ifaceutil.localcpus_for(iface)
    except (OSError, IOError):
        m.warn_msg('Could not retrieve localcpus for interface {0}'
                   .format(iface))
        localcpus = cpui.list_all_cpus()
    ncores, extra_thread = cores
    pairs = cpui.to_siblings_pairs(localcpus)
    good_pairs = sorted(
        filter(cpui.is_not_first_core_of_numa, pairs),
        key=cpui.numa_node_of_pair
    )
    bad_pairs = sorted(
        filter(cpui.is_first_core_of_numa, pairs),
        key=cpui.numa_node_of_pair
    )
    if len(good_pairs) < ncores:
        if not allow_nonlocal:
            m.exit_msg(_not_enough_local_cores_msg(ncores, len(good_pairs)))

        m.warn_msg(_not_enough_local_cores_msg(ncores, len(good_pairs)))
        nonlocal_cpulist = list(set(cpui.list_all_cpus()) - set(localcpus))
        nonlocal_pairs = cpui.to_siblings_pairs(nonlocal_cpulist)
        good_nonlocal_pairs = sorted(
            filter(cpui.is_not_first_core_of_numa, nonlocal_pairs),
            key=cpui.numa_node_of_pair
        )
        good_pairs.extend(good_nonlocal_pairs)
        if len(good_pairs) < ncores:
            m.exit_msg(_not_enough_cores_msg(ncores, len(good_pairs)))

    sp_pairs = good_pairs[0:ncores]

    if not extra_thread:
        return (sp_pairs, None)

    extra_thread_n = bad_pairs[0][1]
    return (sp_pairs, extra_thread_n)


def _get_weights(mtype, iface_acc):
    """Pseudo-weights for the storpool services. Indicate cpu usage."""
    def _sw(i):
        return round(9.998 + (MAX_SERVERS - i - 1) * (10 ** -4), 4)

    old_proto = utils.old_net_proto_compat()

    return dict(
        [
            ('nic', 10),
            ('rdma', 0.9 if iface_acc and not old_proto else 9.999),
            ('iscsi', 9.994),
            ('block', 3.5 if mtype == 'server' else 9),
            ('bridge', 4.6),
            ('mgmt', 4.5),
            ('beacon', 1),
        ]
        + [(s, _sw(i)) for i, s in enumerate(servers.lst(MAX_SERVERS))]
    )


def _add_weights(services, mtype, iface_acc):
    weights = _get_weights(mtype, iface_acc)
    return dict([(s, w) for s, w in weights.items() if s in services])


def gen_service_to_cpu(cpu_to_service):
    """Create the service-cpu dictionary from the cpu-services one."""
    service_to_cpu = collections.defaultdict(list)
    for cpu, services in cpu_to_service.items():
        for service in services:
            service_to_cpu[service].append(cpu)
    return dict(service_to_cpu.items())


def _not_enough_cores_msg():
    msg = 'Allocate more cpu cores to run all services. The current ones ' \
          'are not enough.'
    if not cpuinfo.get().hyperthreading():
        msg += ' Note that the machine does not have hyperthreading.'
    return msg


class CoreServicesConfig(object):
    """Service configuration for a signle core. The core may be hyperthreaded
    (with 2 cpus (default)) or singlethreaded (1 cpu).
    Machine type must be specified.
    """
    def __init__(self, mtype, iface_acc, cpus=2):
        assert cpus in (1, 2)
        self.nthreads = cpus
        self.mtype = mtype
        self.iface_acc = iface_acc
        if cpus == 2:
            self.services = [[], []]
        else:
            self.services = [[], None]

    def assign(self, service, preferred_thread=None):
        """
        Add a service to the core. If a preferred thread is specified,
        the service will be assigned there.
        """
        assert preferred_thread is None or preferred_thread < self.nthreads
        if preferred_thread is not None:
            self.services[preferred_thread].append(service)
            return

        if not self.services[0] or self.nthreads == 1:
            self.services[0].append(service)
        else:
            self.services[1].append(service)

    def remove(self, service):
        """Remove a service from the core."""
        if service in self.services[0]:
            self.services[0].remove(service)
        if self.nthreads == 2 and service in self.services[1]:
            self.services[1].remove(service)

    def has(self, service):
        """Indicates whether the given service is assigned to the core."""
        return service in self.all_services()

    def unused_threads(self):
        """Returns the number of unused threads on the core."""
        unused = 0
        if not self.services[0]:
            unused += 1
        if self.nthreads == 2 and not self.services[1]:
            unused += 1
        return unused

    def thread_of(self, service):
        """
        Returns the thread id of the thread on which the service is assigned.
        """
        for thread_n, thread_services in enumerate(self.services):
            if thread_services is None:
                continue
            if service in thread_services:
                return thread_n
        return None

    def is_empty(self):
        """Indicates whether the core has no services assigned to it."""
        if self.nthreads == 1:
            return not self.services[0]
        return not self.services[0] and not self.services[1]

    def all_services(self):
        """Returns list of all services assigned to the core."""
        srvs = copy.deepcopy(self.services[0])
        if self.nthreads == 2:
            srvs.extend(copy.deepcopy(self.services[1]))
        return srvs

    def max_weight(self):
        """Returns the weight of the heaviest service assgined to the core."""
        srvs = self.all_services()
        if not srvs:
            return 0
        return max(_add_weights(srvs, self.mtype, self.iface_acc).values())

    def weight(self):
        """
        Returns the sum of the weights of the services assigned to the core.
        If the core has only one thread, the result is incremented by 10,
        faking the max weight for the other thread.
        """
        srvs = self.all_services()
        if not srvs:
            return 0
        wgs = sum(_add_weights(srvs, self.mtype, self.iface_acc).values())
        return wgs + 10 if self.nthreads == 1 else wgs

    def __str__(self):
        return str(self.services)


class CPUServiceConfig(object):
    """
    Service configuration for the cpu. Keeps a list of CoreServicesConfig
    object for each core. Initially a signle core is allocated for each service
    and the extra thread(if any) is left unused.
    """
    def __init__(self, services, mtype, iface_acc, extra_thread=False):
        self.mtype = mtype
        self.iface_acc = iface_acc
        self.cpus_per_core = 1 + cpuinfo.get().hyperthreading()
        if self.cpus_per_core == 1:
            assert not extra_thread
        self.cores = [CoreServicesConfig(mtype, iface_acc, self.cpus_per_core)
                      for _ in range(len(services))]
        self.extra = CoreServicesConfig(mtype, iface_acc, 1) if extra_thread \
            else None
        for core, service in zip(self.cores, services):
            core.assign(service)

    def core_of(self, service):
        """
        Return a CoreServicesConfig object which contains the given service.
        """
        for core in self.cores:
            if core.has(service):
                return core
        return None

    def has_extra(self):
        """Indicates whether extra thread is used in the configuration."""
        return self.extra is not None

    def get_extra(self):
        """Returns a CoreServicesConfig object for the extra thread core."""
        return self.extra

    def has(self, service):
        """Indicates whether the given service is assigned to the cpu."""
        return any(core.has(service) for core in self.cores)

    def add_core(self, service):
        """
        Adds a core to the CPUConfig with the given service assigned to it.
        """
        new_core = CoreServicesConfig(
            self.mtype, self.iface_acc, self.cpus_per_core
        )
        new_core.assign(service)
        self.cores.append(new_core)

    def non_empty_cores(self):
        """Returns number of empty cores on the cpu.
        Does NOT count the extra thread core.
        """
        return sum(1 for core in self.cores if not core.is_empty())

    def all_services(self):
        """Returns list of all services assigned to cpus."""
        return utils.list2d_flatten(
            [core.all_services() for core in self.cores]
        )

    def min_weight_non_empty_core(self):
        """Returns the CoreServicesConfig object with the minimum weight."""
        min_core = None
        for core in self.cores:
            if core is None:
                continue
            if not core.is_empty() and \
                    (min_core is None or core.weight() < min_core.weight()):
                min_core = core
        if self.extra is not None:
            if min_core is None or self.extra.weight() < min_core.weight() or \
                    self.extra.is_empty():
                min_core = self.extra
        return min_core

    def __str__(self):
        return '\n'.join(
            [str(core) for core in self.cores] + [str(self.extra)]
        )


class CPUServiceConfigError(Exception):
    """Base error for cpu configuration problems."""


def merge_on_one_core(cpucfg, first, second):
    """
    If the config has first and second services and the cores they are placed
    on have an unused thread, move all services from the second core to the
    first code.
    """
    first_core = cpucfg.core_of(first)
    second_core = cpucfg.core_of(second)
    if first_core is None or second_core is None:
        return
    if first_core.unused_threads() == 1 and second_core.unused_threads() == 1:
        all_second = second_core.all_services()
        for service in all_second:
            second_core.remove(service)
            first_core.assign(service)


def optimize_extra(cpucfg):
    """If the config has extra thread, move all services form mgmt or bridge
    cores to the extra thread.
    """
    extra = cpucfg.get_extra()
    if extra is None:
        return
    for service in ('mgmt', 'bridge'):
        core = cpucfg.core_of(service)
        if core is None:
            continue
        for serv in core.all_services():
            core.remove(serv)
            extra.assign(serv)
        return

    all_serv = cpucfg.all_services()
    if len(all_serv) == 1:
        serv = all_serv[0]
        cpucfg.core_of(serv).remove(serv)
        extra.assign(serv)


def optimize_rdma(cpucfg):
    """Merge a service on the same core with the rdma."""
    rdma_service_merge = (
        ['block', 'iscsi'] + servers.lst(MAX_SERVERS) + ['mgmt', 'bridge']
    )
    for service in rdma_service_merge:
        merge_on_one_core(cpucfg, 'rdma', service)


def add_light(cpucfg, services):
    """Add the given services to the cpu config, sorted by weight and assigned
    to the least loaded core.
    """
    weights = _add_weights(services, cpucfg.mtype, cpucfg.iface_acc)
    for service in sorted(services, key=weights.get, reverse=True):
        cpucfg.min_weight_non_empty_core().assign(service)


def merge_one_thread(cpucfg, first, second):
    """If the config has first and second services, move second service on the
    same core and the same thread as first.
    """
    first_core = cpucfg.core_of(first)
    second_core = cpucfg.core_of(second)
    if first_core is None or second_core is None:
        return
    second_core.remove(second)
    thread = first_core.thread_of(first)
    first_core.assign(second, thread)


def shrink(cpucfg, non_empty_wanted):
    """
    Try to shrink the config until the given number of wanted non empty cores
    is satisfied.
    """
    cpucfg = copy.deepcopy(cpucfg)
    original_cfg = copy.deepcopy(cpucfg)

    def _core_merger_for(merger):
        first, second = merger
        return functools.partial(merge_on_one_core, first=first, second=second)

    def _thread_merger_for(merger):
        first, second = merger
        return functools.partial(merge_one_thread, first=first, second=second)

    srv = servers.lst(MAX_SERVERS)
    bl_srv_list = ['block'] + srv

    merges = [('iscsi', 'mgmt')] + \
        zip(bl_srv_list, itertools.repeat('bridge')) + \
        zip(bl_srv_list, itertools.repeat('mgmt'))

    normal_optimizers = [
        _core_merger_for(('nic', 'rdma')),
        optimize_extra,
        optimize_rdma,
    ] + map(_core_merger_for, merges)

    def normal_optimization(cpucfg):
        """Try some non-radical(basic) optimizations."""
        nopt = 0
        while cpucfg.non_empty_cores() > non_empty_wanted and \
                nopt < len(normal_optimizers):
            optimizer = normal_optimizers[nopt]
            optimizer(cpucfg)
            nopt += 1
        return cpucfg.non_empty_cores() <= non_empty_wanted

    merges = (
        [('iscsi', 'block')]
        + list(itertools.combinations(srv, 2))
        + [('server', 'iscsi')]
    )
    restart_optimizers = (
        map(_core_merger_for, merges)
        + [_thread_merger_for(('mgmt', 'bridge'))]
    )

    ropt = 0
    while cpucfg.non_empty_cores() > non_empty_wanted and \
            ropt <= len(restart_optimizers):
        cpucfg = copy.deepcopy(original_cfg)
        for i in range(ropt):
            optimizer = restart_optimizers[i]
            optimizer(cpucfg)
        ropt += 1
        normal_optimization(cpucfg)
    if cpucfg.non_empty_cores() > non_empty_wanted:
        raise CPUServiceConfigError('Cannot shrink CPU Service configuration '
                                    'to desired number of cores')
    return cpucfg


def make_cpuconfig(services, mtype, iface_acc, ncores, extra_thread):
    """
    Creates a CPUServiceConfig object with the desired configuration
    parameters.
    """
    services = copy.deepcopy(services)
    assert 'beacon' in services
    assert 'rdma' in services
    light = ['beacon']
    services.remove('beacon')
    if mtype == 'server' and block_is_cpu_friendly() and 'block' in services:
        light.append('block')
        services.remove('block')
    if not utils.old_net_proto_compat() and iface_acc:
        light.append('rdma')
        services.remove('rdma')

    if 'bridge' in services and bridge_needs_two_cpus():
        services.remove('bridge')
        cpucfg = CPUServiceConfig(services, mtype, iface_acc, extra_thread)
        cores_to_leave = 1 if cpucfg.cpus_per_core == 2 else 2
        cpucfg = shrink(cpucfg, ncores - cores_to_leave)
        add_light(cpucfg, light)

        cpucfg.add_core('bridge')
        if cores_to_leave == 2:
            cpucfg.add_core('bridge')
        else:
            cpucfg.core_of('bridge').assign('bridge', preferred_thread=1)
    else:
        cpucfg = CPUServiceConfig(services, mtype, iface_acc, extra_thread)
        cpucfg = shrink(cpucfg, ncores)
        add_light(cpucfg, light)
    return cpucfg


def gen_cpu_to_service(services, mtype, iface_acc, sp_pairs, sp_extra_thread):
    """Generates a cpu-services dictionary configuration for the machine."""
    extra = sp_extra_thread is not None
    try:
        cpucfg = make_cpuconfig(services, mtype, iface_acc, len(sp_pairs),
                                extra)
    except CPUServiceConfigError:
        m.exit_msg(_not_enough_cores_msg())
    cpu_to_service = {}
    if extra:
        cpu_to_service[sp_extra_thread] = cpucfg.get_extra().all_services()

    sp_pairs = copy.deepcopy(sp_pairs)
    for core in sorted(cpucfg.cores, key=CoreServicesConfig.max_weight):
        if core.is_empty():
            continue
        cpu0, cpu1 = sp_pairs.pop()
        cpu_to_service[cpu0] = core.services[0]
        if cpu1 is not None:
            cpu_to_service[cpu1] = core.services[1]

    for cpu0, cpu1 in sp_pairs:
        cpu_to_service[cpu0] = cpu_to_service.get(cpu0, [])
        if cpu1 is not None:
            cpu_to_service[cpu1] = cpu_to_service.get(cpu1, [])

    return cpu_to_service


def assign_cpus(cfg):
    """Generates cpu-services configuration and the corresponding service-cpu
    configuration for the machine, obliging the given ControlGroupToolConfig.
    """
    extra = [] if cfg.netconf.iface_acc else ['nic']
    mtype = get_mtype(cfg.serviceconf, cfg.typeconf)
    sp_pairs, sp_extra_thread = get_sp_cpus(cfg.netconf.iface,
                                            cfg.cpuconf.cores_tuple,
                                            cfg.cpuconf.numa_overflow)
    cpu_to_service = gen_cpu_to_service(
        cfg.serviceconf.list_services + extra,
        mtype,
        cfg.netconf.iface_acc,
        sp_pairs,
        sp_extra_thread
    )
    service_to_cpu = gen_service_to_cpu(cpu_to_service)
    if not cfg.netconf.iface_acc:
        m.warn_msg("NIC is expected to be on cpu {0}"
                   .format(service_to_cpu['nic']))
        del service_to_cpu['nic']
    return service_to_cpu, cpu_to_service


def calculate_threads_needed(servconf, netconf, typeconf):
    """Returns the minimum number of threads needed to create a cpu config for
    the machine with the given parameters.
    """
    extra = [] if netconf.iface_acc else ['nic']
    mtype = get_mtype(servconf, typeconf)
    for i in itertools.count():
        cores = i // 2 if cpuinfo.get().hyperthreading() else i
        extra_thread = i % 2 if cpuinfo.get().hyperthreading() else False
        try:
            cpucfg = make_cpuconfig(servconf.list_services + extra, mtype,
                                    netconf.iface_acc, cores, extra_thread)
            threads = cpucfg.cpus_per_core * cpucfg.non_empty_cores()
            if cpucfg.has_extra() and not cpucfg.get_extra().is_empty():
                threads += 1
            return threads
        except CPUServiceConfigError:
            pass
    assert False  # should never be here
