"""Provides different calculations for the storpool memory limits."""

from __future__ import division

from storpool import spconfig

from sp.disk import disklist as spdisk
from sp.util import limits as limutils
from sp.util.misc import get_memtotal_kb

from .config import MAX_SERVERS

DEFAULT_MEM_GB = {
    'mgmt': 2,
    'bridge': 1,
    'iscsi': 1
}


HDD_NEEDS_GB_FOR_TB = 0.28
SSD_NEEDS_GB_FOR_TB = 0.48
ADDED_MARGIN_MULT = 0.2  # 20%
MIN_SERVER_MEM_GB = 2
GB_SPACE_PER_DISK = 0.5


CACHE_METADADA_MB_PER_GB = 44


def size_with_metadata(cache_size_gb):
    """Adds metadata size to the usable memory size of a shared memory file."""
    return cache_size_gb + (cache_size_gb * CACHE_METADADA_MB_PER_GB) / 1024


def servers_needs():
    """Returns a pair of lists - cache memory needs for each server and
    general memory needs for each server. All values are in GB.
    """
    instances = spdisk.server_instances()
    disks = spdisk.all_disks_with_caps()
    hddcap = 0
    ssdcap = 0

    disks_per_instance = [0] * instances
    for disk in disks:
        if 'journal' in disk['meta']:
            continue
        if 'SSD' in disk['meta']:
            ssdcap += disk['cap']
        else:
            hddcap += disk['cap']
        disks_per_instance[disk['meta']['instance']] += 1

    cache_needs = [ndisks * GB_SPACE_PER_DISK for ndisks in disks_per_instance]
    tbhdd = int(hddcap) / 1024 ** 4
    tbssd = int(ssdcap) / 1024 ** 4
    needs = HDD_NEEDS_GB_FOR_TB * tbhdd + SSD_NEEDS_GB_FOR_TB * tbssd
    needs += ADDED_MARGIN_MULT * needs
    # guarantee min memory for each instance
    needs = max(needs, instances * MIN_SERVER_MEM_GB)
    return cache_needs, needs


def cache_sizes_from_conf(instances=None):
    """Returns list of server cache sizes(in GB) from configuration.
    Number of server instances may be specified.
    """
    if instances is None:
        instances = spdisk.server_instances()
    spcfg = spconfig.SPConfig()
    default_cache = int(spcfg['SP_CACHE_SIZE'])
    cache_needs = [int(spcfg.get('SP_CACHE_SIZE_{0}'.format(i), default_cache))
                   for i in range(instances)]
    return [cache_size / 1024 for cache_size in cache_needs]


SERVICE_LOG_FILE_SIZE_GB = 128 / 1024


def sp_alloc_limit(iscsi, mgmt, bridge):
    """Returns memory limit string for storpool.slice/alloc"""
    limgb = 0
    limgb += DEFAULT_MEM_GB['iscsi'] if iscsi else 0
    limgb += DEFAULT_MEM_GB['mgmt'] if mgmt else 0
    limgb += DEFAULT_MEM_GB['bridge'] if bridge else 0
    log_files_count = iscsi + mgmt + bridge
    limgb += log_files_count * SERVICE_LOG_FILE_SIZE_GB
    return limutils.limit_to_gb_str(limgb)


CACHE_SIZE_TMPL = "SP_CACHE_SIZE_{n}={s}"


def get_cache_conf(calculate=False):
    """Generates the content of the cache.conf file."""
    if calculate:
        cache_needs, _ = servers_needs()
    else:
        cache_needs = cache_sizes_from_conf()
    cache_confs = [CACHE_SIZE_TMPL.format(n=n, s=int(size * 1024))
                   for n, size in zip(range(MAX_SERVERS), cache_needs)]
    return '\n'.join(cache_confs)


def _sp_common_log_files_needs(servconf):
    log_files_count = servconf.servers + servconf.block
    return size_with_metadata(log_files_count * SERVICE_LOG_FILE_SIZE_GB)


def _sp_common_block_needs():
    bytes_per_4k_page = 8
    pages = get_memtotal_kb() / 4
    bytes_needed = bytes_per_4k_page * pages
    return (bytes_needed / 1024 ** 3) + 0.3  # add 300 mb margin


def _sp_common_server_needs(memconf):
    cache_needs, general_server_needs = servers_needs()
    if not memconf.set_cache_size:
        cache_needs = cache_sizes_from_conf()
    return general_server_needs + sum(map(size_with_metadata, cache_needs))


def sp_common_min_limit(memconf, servconf):
    """Returns minimum memory limit string for storpool.slice/common"""
    limitgb = _sp_common_log_files_needs(servconf)
    if servconf.block:
        limitgb += _sp_common_block_needs()
    if servconf.servers:
        limitgb += _sp_common_server_needs(memconf)
    return limutils.limit_to_gb_str(limitgb)


CTRL_BYTES_PER_ENTRY = 112


def controller_peak_usage_mb():
    """
    Returns the peak memory usage of the storpool_controller for a disk(in MB).
    """
    disks = spdisk.all_disks_with_caps()
    if not disks:
        return 0
    max_entries = max(spdisk.entries_count(disk) for disk in disks)
    peak_bytes = CTRL_BYTES_PER_ENTRY * max_entries
    return peak_bytes // 1024 ** 2


MIN_CTRL_PEAK_TO_ADD = 200


def system_limit_bump():
    """Calculates the limit bump for the system slice, depending on
    storpool_controller peak memory usage. Return limit string.
    """
    ctrl_peak = controller_peak_usage_mb()
    if ctrl_peak < MIN_CTRL_PEAK_TO_ADD:
        return '0M'
    return '{0}M'.format(ctrl_peak)
