"""Storpool process."""
from __future__ import print_function

import argparse

from sp import messages as m
from sp.cgroups import hierarchy
from sp.process import sppidutil


COLUMNS = {
    'pid': {'length': 5, 'number': 0},
    'service': {'length': 10, 'number': 1},
    'cpuset': {'length': 23, 'number': 2},
    'memory': {'length': 21, 'number': 3},
}

COLUMN_NAMES = sorted(COLUMNS.keys(), key=lambda k: COLUMNS[k]['number'])


def get_args_parser():
    """Return command line arguments parser."""
    parser = argparse.ArgumentParser(description='StorPool processes utility.')

    subparsers = parser.add_subparsers(help='command', dest='command')

    printp = subparsers.add_parser('list')
    printp.add_argument('-C', '--columns', nargs='+', help='columns to show',
                        choices=COLUMN_NAMES)
    printp.add_argument('-S', '--sort', nargs='+',
                        help='sort keys(column names)',
                        choices=COLUMN_NAMES)
    printp.add_argument('-H', '--no-header', action='store_true',
                        default=False,
                        help='do not show column names as a table header in '
                             'the output')

    reclsp = subparsers.add_parser('reclassify')
    reclsp.add_argument('-N', '--noop', action='store_true', default=False,
                        help='no operation')
    reclsp.add_argument('-v', '--verbose', action='count', default=0,
                        help='verbose level')
    return parser


def print_info(info, columns=COLUMN_NAMES):
    """Print a sp process info object with the desired columns."""
    nfo = dict([(col_name, info[col_info['number']])
                for col_name, col_info in COLUMNS.items()])
    for column in columns:
        tmpl = '{{0:<{length}}} '.format(length=COLUMNS[column]['length'])
        print(tmpl.format(nfo[column]), end='')
    print()


def update_column_lengths(nfos, no_header=False):
    """Update the default column lengths, based on the coming output."""
    if not no_header:
        nfos = nfos + [tuple('[{0}]'.format(cname) for cname in COLUMN_NAMES)]
    for info in COLUMNS.values():
        info['length'] = max(len(str(nfo[info['number']])) for nfo in nfos)


def reclassify(hierarchy_controller):
    """Classify all sp processes to the configured cgroups."""
    def _classify(controller, group, pid):
        try:
            hierarchy_controller.safe_classify(controller, group, pid)
        except (OSError, ValueError) as err:
            m.err_msg('Could not move pid {pid} to {ctrl}/{grp}'
                      .format(pid=pid, ctrl=controller, grp=group))
            m.err_msg(err)

    config = sppidutil.sp_cg_config()

    def _reclassify_one(nfo):
        pid, service, cpuset, memory = nfo
        cpu_ok = cpuset.startswith(config[service]['cpuset'])
        mem_ok = memory.startswith(config[service]['memory'])
        not_ok_tmpl = '{pid}({service}) {cgroup} group is {crr_group} ' \
                      'while the configured one is {conf_group}'
        if not cpu_ok:
            m.msg(not_ok_tmpl.format(pid=pid,
                                     service=service,
                                     cgroup='cpuset',
                                     crr_group=cpuset,
                                     conf_group=config[service]['cpuset']))
            _classify('cpuset', config[service]['cpuset'], pid)
        if not mem_ok:
            m.msg(not_ok_tmpl.format(pid=pid,
                                     service=service,
                                     cgroup='memory',
                                     crr_group=memory,
                                     conf_group=config[service]['memory']))
            _classify('memory', config[service]['memory'], pid)

    map(_reclassify_one, sppidutil.sp_pids_info())


def main():
    """Parse the command line arguments and execute the required actions."""
    args = get_args_parser().parse_args()
    if args.command == 'reclassify':
        m.set_verbose_level(args.verbose)
        reclassify(hierarchy.get(mock=args.noop))
    else:
        columns = args.columns if args.columns else COLUMN_NAMES
        nfos = sppidutil.sp_pids_info()
        update_column_lengths(nfos, args.no_header)
        if not args.no_header and nfos:
            print_info(['[{0}]'.format(col) for col in COLUMN_NAMES], columns)
        sort = [COLUMNS[col]['number'] for col in args.sort] if args.sort else []
        sort = sort + [p for p in range(4) if p not in sort]

        for nfo in sorted(nfos, key=lambda nfo: tuple(nfo[i] for i in sort)):
            print_info(nfo, columns)
