#!/usr/bin/env python

'''
Used to help splitting already existing disks into multiple
storpool_server instances, by printing required storpoot_initdisk commands
'''
import argparse
import os
import sys

# pylint: disable=wrong-import-position
sys.path.append('/usr/lib/storpool/python')
from sp.util.backports import subprocess

class StorPoolDisk(object):
    """StorPoolDisk object
    example 'splitline' output from `storpool_initdisk --li`
    ['/dev/sdc1', ' diskId 1688', ' version 10007', ' server instance 0', \
    ' cluster a.o', ' WBC']
    properties
    self.line - the above split line from `storpool_initdisk --list`
    self.blockp - path to block device
    self.did - diskId
    self.instance - server instance
    self.cluster - cluster ID
    self.flags - disk flags
    """
    def __init__(self, line):
        super(StorPoolDisk, self).__init__()
        self.line = line
        self.blockp = line[0]
        self.did = line[1].split()[-1]
        version = int(line[2].split()[-1])
        if version == 10006:
            self.instance = self.cluster = None
            self.flags = [i.strip() for i in line[3:]]
        elif version in (10007, 10009):
            self.instance = int(line[3].split()[-1])
            self.cluster = line[4].split()[-1]
            self.flags = [i.strip() for i in line[5:]]
        else:
            version = None
            self.instance = self.cluster = self.flags = None

    def __repr__(self):
        '''
        returns string that would create  the same object.
        '''
        return 'StorPoolDisk({ln})'.format(ln=self.line)

    def __str__(self):
        '''
        shows all variables returned by __init__ in a tuple
        '''
        return str(
            (
                self.blockp,
                self.did,
                self.instance,
                self.cluster,
                self.flags,
                self.ssd
            )
        )

    @property
    def ssd(self):
        """Property showing whether the disk is an SSD"""
        return self.flags and 'SSD' in self.flags

    @property
    def journal(self):
        """Property showing whether the disk is a journal"""
        return self.flags and [i for i in self.flags if 'journal' in i]

    @property
    def jmv(self):
        """Property showing whether the disk has a journal"""
        res = None
        for i in self.flags:
            if 'mv' in i or 'jmv' in i:
                res = i.split()[-1]
        return res

    @property
    def journaldev(self):
        """Property showing the journaldev for the disk"""
        alldisks = const()
        for disk in alldisks:
            try:
                if self.jmv in disk.flags[-1] and 'journal' in disk.flags[-1]:
                    return disk
            except IndexError:
                pass
        return None

    @property
    def ejected(self):
        """Property showing the eject flag status for the disk"""
        return self.flags and 'EJECTED' in self.flags


def which(program):
    '''program: str name of the executable
    finds an executable, returns path to it if any'''
    def is_exe(fpath):
        '''fpath: str path to file, chk if executable'''
        return os.path.isfile(fpath) and os.access(fpath, os.X_OK)
    fpath, _ = os.path.split(program)
    if fpath:
        if is_exe(program):
            return program
    else:
        for path in os.environ["PATH"].split(os.pathsep):
            path = path.strip('"')
            exe_file = os.path.join(path, program)
            if is_exe(exe_file):
                return exe_file
    return None


def const():
    '''
    returns list with StorPoolDisk objects constructed from
    `storpool_initdisk --list`
    '''
    # get the output
    result = []
    try:
        cmd = '{i} --list'.format(i=INITDISK).split()
        alllines = subprocess.check_output(cmd, stderr=None)
    except subprocess.CalledProcessError as err:
        exit(err)
    for line in alllines.split('\n'):
        spl = line.split(',')
        if len(spl) == 1:
            if spl[0] == '' or spl[0] == 'Done.':
                continue
        result.append(StorPoolDisk(spl))
    return result


def chunks(lst, num):
    '''
    Yield successive num-sized chunks from lst
    '''
    for i in range(0, len(lst), num):
        yield lst[i:i + num]


def commands(disks, instances, mod=None):
    '''
    disks: list of StorPoolDisk objects
    instances: int - number of desired server instances
    prints commands needed to disperse over instances
    '''
    def pcmds(slices, inst, mod=None):
        """Prints final commands"""
        disks = slices.pop()
        if mod:
            inst += mod
        for disk in disks:
            if disk.journal:
                continue  # never print a journal
            jarg = ''
            if disk.jmv:
                jarg = '--journal {j}'.format(j=disk.journaldev.blockp)
            print(
                '{s} -r -i {i} {did} {b} {j} # {ssd}'.format(
                    s=INITDISK,
                    i=inst,
                    did=disk.did,
                    b=disk.blockp,
                    ssd=' '.join(disk.flags),
                    j=jarg
                )
            )
    nslices = len(disks)/instances
    slices = [i for i in chunks(disks, nslices)]
    for inst in range(instances):
        pcmds(slices, inst, mod)
    for inst in range(len(slices)):
        pcmds(slices, inst, mod)


def getargs():
    '''returns argparse instance object'''
    parser = argparse.ArgumentParser(description='''
    Prints relevant commands for dispersing the drives to multiple
    server instances
    ''')
    parser.add_argument(
        '-i',
        '--instances',
        type=int,
        help="Number of instances",
        default=2
    )
    parser.add_argument(
        '-s',
        '--ssd-only',
        help=(
            "Splits by type, 's' SSD-only instances plus"
            "i-s HDD instances (default s: 1)"
        ),
        nargs='?',
        const=1,
        type=int,
    )
    return parser.parse_args()


def main():
    '''
    instances: int - number of instances
    '''
    args = getargs()
    if args.instances < 1 or args.instances > DEFMAXINST:
        sys.exit(
            'Present max number of instances 1-{u}'.format(
                u=DEFMAXINST
            )
        )
    disks = const()
    alldisks = [d for d in disks if not d.journal and not d.ejected]
    if len(alldisks) < args.instances:
        sys.exit(
            'Need at least {i} disks, have {d}'.format(
                i=args.instances,
                d=len(alldisks)
            )
        )
    elif len(alldisks) == args.instances:
        commands(alldisks, args.instances)
    else:
        ssds = [s for s in alldisks if s.ssd]
        hdds = [h for h in alldisks if not h.ssd]
        if args.ssd_only:
            if args.ssd_only >= args.instances and hdds:
                sys.exit(
                    '--ssd-only should be less than all instances when there'
                    ' are hard drives available, SSD/all: {s}/{i}'.format(
                        s=args.ssd_only,
                        i=args.instances
                    )
                )
            commands(ssds, args.ssd_only)
            commands(hdds, args.instances - args.ssd_only, args.ssd_only)
        else:
            if len(ssds) < args.instances or len(hdds) < args.instances:
                commands(alldisks, args.instances)
            else:
                commands(ssds, args.instances)
                commands(hdds, args.instances)


if __name__ == '__main__':
    try:
        INITDISK = which('storpool_initdisk')
        DEFMAXINST = 7
        main()
    except KeyboardInterrupt as err:
        exit(err)
