#!/usr/bin/python

from __future__ import print_function
import argparse
import json
import subprocess
import sys
import storpool.spapi

def getargs():
    """Parse arguments, returns argparse.ArgumentParser object instance"""
    parser = argparse.ArgumentParser(
        description="""
        To be used for the upgrade from 18.02 to 19.01 release

        - dumps data for exports, locations, clusters and remote bridges
        in json format
        - cleans up all remote bridge info from dump file
        - restores remote bridges from file
        """
    )
    parser.add_argument(
        '-d',
        '--dump',
        type=str,
        help="Path to file to dump snapshot exports and remote data into"
    )
    parser.add_argument(
        '-c',
        '--cleanup',
        type=str,
        help="Path to json data to cleanup (exports, clusters, locations)"
    )
    parser.add_argument(
        '-r',
        '--restore',
        type=str,
        help="Path to file to restore snapshot exports and remote bridges from"
    )
    return parser.parse_args()


def cmdexec(cmd):
    """
    cmd: list of str args
    returns output from command, exits in case of an error detected"""
    subproc = subprocess.Popen(
        cmd,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE
    )
    stdout, stderr = subproc.communicate()
    result = subproc.wait()
    if result != 0:
        sys.exit(
            "Command failed (exit {e}):\n{c}\nstderr:\n{s}\n".format(
                e=result, c=" ".join(cmd), s=stderr
            )
        )
    return stdout


def dump(ftarget, api):
    """
    ftarget: str path to file
    api: storpool.spapi.Api instance object
    collects snapshot exports, remote locations, clusters and bridges
    dumps them into ftarget as json string"""
    result = {}
    result['exports'] = [
        i.toJson() for i in api.exportsList().get('exports', [])
    ]
    result['locations'] = [
        i.toJson() for i in api.locationsList().get('locations', [])
    ]
    clusters = json.loads(cmdexec(['storpool', '-Bj', 'cluster', 'list']))
    result.update(clusters.get('data', {'clusters': []}))
    remotebridges = json.loads(
        cmdexec(['storpool', '-Bj', 'remoteBridge', 'list'])
    )
    result.update(remotebridges.get('data', {'remoteBridges': []}))
    with open(ftarget, 'w+') as ftgt:
        json.dump(result, ftgt, indent=True)
    if readjson(ftarget) != result:
        sys.exit('Differences found between data and file, bailing out')

def readjson(ftarget):
    """
    ftarget: str path to file
    """
    with open(ftarget) as ftgt:
        return json.load(ftgt)


def validaterequiredkeys(bridge):
    """bridge: dict with bridge data
    ensures required data is present, exit in case it is not
    """
    missingkeys = []
    for key in ['ip', 'location', 'publicKey']:
        if key not in bridge:
            missingkeys.append(key)
    if missingkeys:
        sys.exit(
            'Refusing to cleanup, required keys missing:\n{k}'.format(
                k=", ".join(missingkeys)
            )
        )


def cleanup(ftarget):
    """Reads data from ftarget json file and
    - removes all defined 'remoteBridges'
    """
    data = readjson(ftarget).get('remoteBridges', [])
    if not data:
        sys.exit('No remote bridge info in {f}'.format(f=ftarget))
    for remotebridge in data:
        validaterequiredkeys(remotebridge)
        cmd = [
            'storpool',
            '-B',
            'remoteBridge',
            'deregister',
            remotebridge.get('ip')
        ]
        print(cmdexec(cmd))


def restore(ftarget):
    """Reads data from ftarget json file and
    - creates all defined 'locations'
    - creates all defined 'clusters'
    - exports all 'exports' snapshots objects
    """
    data = readjson(ftarget).get('remoteBridges', [])
    if not data:
        sys.exit('No remote bridge info in {f}'.format(f=ftarget))
    for remotebridge in data:
        validaterequiredkeys(remotebridge)
        cmd = [
            'storpool',
            '-B',
            'remoteBridge',
            'register',
            remotebridge.get('location'),
            remotebridge.get('ip'),
            remotebridge.get('publicKey')
        ]
        deletedelay = remotebridge.get('minimumDeleteDelay')
        if deletedelay:
            cmd.append('{i}s'.format(i=deletedelay))
        if remotebridge.get('noCrypto'):
            cmd.append('noCrypto')
        print(cmdexec(cmd))


def main():
    """
    Runs checks, then either dumps or restores configuration from json file
    """
    args = getargs()
    if not any([args.dump, args.cleanup, args.restore]):
        sys.exit('Please select either --dump or --restore')
    api = storpool.spapi.Api.fromConfig()
    if args.dump:
        dump(args.dump, api)
    elif args.cleanup:
        cleanup(args.cleanup)
    elif args.restore:
        restore(args.restore)
    # elif args.restore:
    #     pass

if __name__ == '__main__':
    try:
        main()
    except KeyboardInterrupt:
        sys.exit('\nKeyboardInterrupt detected, bailing out')
