"""Provides functions for easy access to common process properties."""

import os
import re


from sp.util import re_extensions as extre
from sp.util.file import readers


class ProcfsError(Exception):
    """An error that occured while reading the proc filesystem."""


class StatFileError(ProcfsError):
    """An error that occured while reading a stat file in the proc filesystem.
    """


class CommFileError(ProcfsError):
    """An error that occured while reading a comm file in the proc filesystem.
    """


def get_comm(pid):
    """Returns the comm of the given pid."""
    comm_name = '/proc/{pid}/comm'.format(pid=pid)
    try:
        with open(comm_name, mode='rb') as comm_file:
            data = comm_file.read()
            if not data:
                raise CommFileError('The {comm} file seems to be empty'
                                    .format(comm=comm_name))
            elif data[-1:] != '\n'.encode('us-ascii'):
                raise CommFileError('The {comm} file does not end with a newline'
                                    .format(comm=comm_name))
            comm = data[:-1]
    except (IOError, OSError) as err:
        raise CommFileError('Could not read the {comm} file: {err}'
                            .format(comm=comm_name, err=err))
    return comm


def get_stat(pid):
    """Returns the stat of the given pid."""
    comm = get_comm(pid)
    prefix = '{pid} ('.format(pid=pid).encode('us-ascii') + \
             comm + \
             ') '.encode('us-ascii')
    stat_name = '/proc/{pid}/stat'.format(pid=pid)
    try:
        with open(stat_name, mode='rb') as stat_file:
            data = stat_file.read()
            if len(data) < len(prefix):
                raise StatFileError('The {stat} file is less than {cnt} bytes long'
                                    .format(stat=stat_name, cnt=len(prefix)))
            elif data[0:len(prefix)] != prefix:
                raise StatFileError('The {stat} file does not start correctly'
                                    .format(stat=stat_name))

            # Let us hope that the rest of the data is only numbers, mmkay?
            stat_fields = data[len(prefix):].decode('us-ascii').split()
    except (IOError, OSError) as err:
        raise StatFileError('Could not read the {stat} file: {err}'
                            .format(stat=stat_name, err=err))

    return [pid, comm] + stat_fields


def get_parent(pid):
    """Returns the ppid of the given pid."""
    return int(get_stat(pid)[3])


def get_status(pid):
    """Returns the status of the given pid."""
    return get_stat(pid)[2]


TGID_RE = re.compile(r"Tgid:\s+(?P<tgid>\d+)")


def get_tgid(pid):
    """Returns the tgid of the given pid."""
    tgid_matches = extre.matches(TGID_RE, readers.lines('/proc/{0}/status'.format(pid)))
    if not tgid_matches:
        return None
    return int(tgid_matches[0].group('tgid'))


def get_cmdline(pid):
    """Returns the cmdline of the given pid."""
    return readers.lines('/proc/{0}/cmdline'.format(pid))[0]


def exists(pid):
    """Indicates whether the given pid exists."""
    return os.path.isdir('/proc/{0}'.format(pid))


def cgroups(pid):
    """Returns a cgroups controller-cgroup dictionary for the given pid."""
    raw = readers.lines('/proc/{0}/cgroup'.format(pid))
    return dict([line.rstrip().split(':')[1:3] for line in raw])


def cgroup(pid, controller):
    """Returns the cgroup for the given controller of the given pid."""
    group = cgroups(pid)[controller]
    return '.' if group == '/' else group[1:]


def ancestors_of(pid):
    """Returns list of acestors(pids) of the given pid in the process tree."""
    ancestors = []
    while get_parent(pid) != 0:
        ancestors.append(get_parent(pid))
        pid = get_parent(pid)
    return ancestors


def all_pids(no_zombies=True):
    """Returns a list of all pids.
    If no_zombies parameter is set to True, zombie processes will be filtered.
    """
    def _alive_not_zombie(pid):
        try:
            return get_status(pid) != 'Z'
        except ProcfsError:
            return False

    pids = [int(d) for d in os.listdir('/proc') if d.isdigit()]
    if no_zombies:
        pids = [pid for pid in pids if _alive_not_zombie(pid)]
    return pids
