'''
Sends operational data collected from services for monitoring and analysis
'''

from __future__ import print_function

import argparse
import errno
import fcntl
import fnmatch
import glob
import itertools
import logging
import logging.handlers
import os
try:
    from pxssh import pxssh, TIMEOUT, EOF, ExceptionPxssh
except ImportError:
    from pexpect.pxssh import pxssh as pxssh
    from pexpect.pxssh import ExceptionPxssh
    from pexpect import TIMEOUT, EOF
import platform
import pwd
import re
import shutil
import subprocess
import sys
import tempfile
import time
import traceback
from collections import defaultdict
from distutils.version import LooseVersion
from socket import gethostname

import storpool.spconfig

from sp.util.file import readers, writers


class AbrtsyncConfig(object):
    # pylint: disable=too-few-public-methods
    '''Just holds some re-usable configuration'''
    args = None
    basedir = None
    ppid = 0
    SP_RAUTH_TOKEN = None
    SP_DELETE_REPORTS = None
    SP_ABRTSYNC_USER = None
    SP_CLUSTER_NAME = None
    STRBOOL = ['1', 'y', 'yes', 't', 'true']
    EMPTY = [None, '']
    logger = None
    VARCRASH = None
    reportdir = None
    sentfiles = defaultdict(bool)
    testfile = None

    @property
    def spconf(self):
        '''Returns new instance on each invocation'''
        return storpool.spconfig.SPConfig()


def get_logger():
    '''Attempts to initialize and return logger instance'''
    try:
        logger = logging.getLogger(sys.argv[0])
        logger.setLevel(logging.INFO)
        formatter = logging.Formatter('%(name)s[%(process)d]: %(message)s')
        dhandler = logging.handlers.SysLogHandler(address='/dev/log')
        dhandler.setFormatter(formatter)
        logger.addHandler(dhandler)
    except Exception as err:  # pylint: disable=broad-except
        print(
            "Unable to instantiate logger, bailing out: {e}".format(e=err),
            file=sys.stderr
        )
        traceback.print_stack(file=sys.stderr)
        sys.exit(1)
    return logger


def getvarcrash():
    '''
    Checks for kdump config, returns str path to directory with kdump
    collected vmcore files
    '''
    kdumpconf = "/etc/kdump.conf"
    kdumptools = "/etc/default/kdump-tools"
    # CentOS like
    if os.path.isfile(kdumpconf):
        for line in open(kdumpconf).readlines():
            if re.match("^path ", line):
                varcrashdir = line.split()[-1].rstrip("\n")
                return varcrashdir
    # Debian like
    elif os.path.isfile(kdumptools):
        for line in open(kdumptools):
            if re.match("KDUMP_COREDIR", line):
                varcrashdir = line.split('=')[-1].rstrip("\n")
                varcrashdir = re.sub("[\"\']+", '', varcrashdir)
                return varcrashdir
    else:
        getconf().logger.error(
            (
                "Could not find kdump crash directory from either "
                "%s, nor %s, is kdump-tools installed?"
            ),
            kdumpconf,
            kdumptools
        )
    return None


def init():
    '''
    used to initialize directories and global variables
    switches as the SP_ABRTSYNC_USER if not the current user
    returns None
    '''
    cfg = getconf()

    cfg.SP_CLUSTER_NAME = cfg.spconf.get('SP_CLUSTER_NAME', None)
    if cfg.SP_CLUSTER_NAME in cfg.EMPTY:
        sys.exit(
            'SP_CLUSTER_NAME does not seem to be configured, please check...'
        )
    cfg.SP_RAUTH_TOKEN = cfg.spconf.get('SP_RAUTH_TOKEN', None)
    if cfg.SP_RAUTH_TOKEN in cfg.EMPTY:
        sys.exit(
            'SP_RAUTH_TOKEN does not seem to be configured, please check...'
        )
    delreports = cfg.spconf.get('SP_DELETE_REPORTS', None)
    if delreports in cfg.STRBOOL:
        cfg.logger.info(
            'SP_DELETE_REPORTS is %s, cleaning up upon successfull send',
            delreports
        )
        cfg.SP_DELETE_REPORTS = True
    # ensure SP_ABRTSYNC_USER is configured
    cfg.SP_ABRTSYNC_USER = cfg.spconf.get('SP_ABRTSYNC_USER', None)
    if cfg.SP_ABRTSYNC_USER in cfg.EMPTY:
        sys.exit('SP_ABRTSYNC_USER not set, bailing out.')
    try:
        pwrec = pwd.getpwnam(cfg.SP_ABRTSYNC_USER)
        uname = pwrec.pw_name
        udir = pwrec.pw_dir
        uuid = pwrec.pw_uid
        ugid = pwrec.pw_gid
    except KeyError:
        sys.exit('{u} does not exist, exiting.'.format(u=cfg.SP_ABRTSYNC_USER))
    # ensure SP_ABRTSYNC_USER home exists
    if not os.path.isdir(udir):
        sys.exit('{u} home directory, does not exist, exiting.'.format(u=udir))
    env = os.environ.copy()
    if env.get('USER') != cfg.SP_ABRTSYNC_USER:
        cfg.logger.info(
            'Switching to unprivileged user "%s" (initiated by "%s")',
            cfg.SP_ABRTSYNC_USER,
            env.get('USER')
        )
        os.setgid(ugid)
        os.setuid(uuid)
        # the following environment change affects child processes
        # when daemonized
        os.environ['LOGNAME'] = uname
        os.environ['HOME'] = udir
        os.environ['PWD'] = udir
        os.environ['USER'] = uname
    name = os.path.basename(sys.argv[0]).lstrip('storpool_').rstrip('.bin')
    cfg.basedir = "{h}/{s}".format(h=udir, s=name)
    cfg.testfile = os.path.join(cfg.basedir, 'abrtsync-testfile')
    if os.path.exists(cfg.testfile):
        os.unlink(cfg.testfile)
    cfg.logger.info('basedir is %s', cfg.basedir)
    if not os.path.isdir(cfg.basedir):
        cfg.logger.info('%s does not exist, creating...', cfg.basedir)
        os.mkdir(cfg.basedir)
    os.chdir(os.environ['PWD'])
    cfg.VARCRASH = getvarcrash()
    cfg.ppid = os.getpid()


def getargs():
    '''collects arguments for this tool, returns ArgumentParser instance'''
    parser = argparse.ArgumentParser(description='''
    Checks for new debug data from past vmcore files (usually in /var/crash)
    and for new report files with debug data from aborted StorPool services
    (default in SP_REPORTDIR).
    Executes `/usr/lib/storpool/storpool_kdump` with root privileges, then
    forks as the SP_ABRTSYNC_USER unprivileged user (default 'storpool').
    ''')
    parser.add_argument(
        '-d',
        '--daemon',
        help=(
            "Daemonize. Runs as a background service regularly "
            "checking/sending new files in SP_REPORTDIR"
        ),
        action='store_true'
    )
    parser.add_argument(
        '-N',
        '--noop',
        help="Run just once (useful for debugging)",
        action='store_true'
    )
    parser.add_argument(
        '-l',
        '--syslog',
        action='store_true',
        default=True,
        help='Log through syslog (ignored, always true)'
    )
    parser.add_argument(
        '-p',
        '--pidfile',
        type=str,
        default='/var/run/storpool/storpool_abrtsync.bin.pid',
        help='The full path of the PID file to store'
    )
    parser.add_argument(
        '-W',
        '--watch-fd',
        type=int,
        help=(
            'The file descriptor used for tracking the connection to '
            'storpool_daemon'
        )
    )
    return parser.parse_args()


def getpresentkernel():
    '''Returns the presently running kernel'''
    plu = platform.uname()
    if plu[0] == 'Linux':
        return unicode(plu[2])
    else:
        sys.exit(
            'Unsupported system, expected Linux, got {p}'.format(p=plu[0])
        )


def getunsupportedkernels(kdir='/lib/modules', bdir='/boot'):
    '''
    checks for all installed kernels that does not have the
    /lib/modules/KERNELVER/extra/storpool_*.ko files
    returns
    (string presently running kernel, \
    [list of string kernels sorted by version])'''
    rstring = '^vmlinu[zx]-'
    kreg = re.compile(rstring)
    kernels = []
    for (_, _, files) in os.walk(bdir):
        for kfile in files:
            if re.match(kreg, kfile):
                # /boot/vmlinuz-4.4.0-101-generic
                # /boot/vmlinuz-2.6.32-696.13.2.el6.x86_64
                kernels.append(re.sub(rstring, '', kfile))
    # sort all kernels (removing the el.* at the end for proper version sort)
    kernels = sorted(
        kernels, key=(lambda x: LooseVersion(re.sub(r'\.el[67].*$', '', x)))
    )
    # consider only kernels after the presently running
    presentkernel = getpresentkernel()
    # return only kernels that are:
    # - later than the presently running one
    kernelstocheck = [
        k for k in kernels[(kernels.index(presentkernel)+1):]
        if os.path.isdir(os.path.join(kdir, k))
    ]
    # - do have a /lib/moduleskernelname directory and does not have
    # extra/storpool_*.ko in their /lib/modules/<kernel>
    unsupportedkernels = [
        unicode(u) for u in kernelstocheck
        if os.path.isdir(os.path.join(kdir, u, 'kernel')) and
        not glob.glob(os.path.join(kdir, u, 'extra/storpool_*.ko'))
    ]
    return unsupportedkernels


def getcputype():
    ''' check for processor type and returns string cputype'''
    with open('/proc/cpuinfo') as cpuf:
        data = cpuf.read()
        result = u'unsupported'
        if re.search(re.compile('AuthenticAMD'), data):
            if re.search(re.compile('epyc', re.I), data):
                result = u'nehalem'
            else:
                result = u'amdfam10'
        elif re.search(re.compile('Atom'), data):
            result = u'atom'
        elif re.search(re.compile("X3220|Q6600"), data):
            result = u'core2'
        elif re.search(re.compile("POWER8"), data):
            result = u'power8'
        elif re.search(re.compile('Intel'), data):
            result = u'nehalem'
        return result


def getos():
    ''' check for running OS and return string os str'''
    osrel = '/etc/os-release'
    rhrel = '/etc/redhat-release'
    unk = u'unknown'
    cfg = {}
    if os.path.isfile(osrel):
        with open(osrel) as mft:
            mdata = mft.read()
            for spl in mdata.split('\n'):
                if len(spl) > 1:
                    key, val = spl.split('=')
                    cfg[key] = val.strip('"')
            osid = cfg.get('ID')
            ver = cfg.get('VERSION_ID')
            if osid and ver:
                return u'{s}_{v}'.format(s=osid, v=ver)
    elif os.path.isfile(rhrel):
        with open(rhrel) as mft:
            mdata = mft.read()
            if re.search(re.compile('(?i)centos release 6'), mdata):
                return u'centos_6'
    return unk


def getrevision():
    """Return contents of /etc/storpool_revision"""
    return readers.getdataf("/etc/storpool_revision").rstrip('\n')


def getdataflines(ftarget):
    '''returns lines from ftarget (consider reusing the readers here)'''
    result = []
    try:
        result.extend(readers.getdataf(ftarget).split('\n'))
    except AttributeError:
        pass
    return result


def touch(path):
    '''Creates a file at path if it does not exist'''
    with open(path, 'a'):
        os.utime(path, None)


def getinventory():
    """collects information regarding cpu, kernel, Operating System and
    installel packages (FIXME) and returns dict object with all the info"""
    inventorydata = {u'data': {}}
    inventorydata[u'data'][u'cputype'] = getcputype()
    inventorydata[u'data'][u'os'] = getos()
    inventorydata[u'data'][u'kernel'] = getpresentkernel()
    inventorydata[u'data'][u'unsupportedkernels'] = getunsupportedkernels()
    inventorydata[u'data'][u'revision'] = getrevision()
    return inventorydata


def configureremote(remote):
    '''Attempts to configure a remote with public key'''
    cfg = getconf()
    # first ensure if it wasn't configured already
    if testremote(remote):
        return True
    keyfile = os.path.join(os.environ['HOME'], '.ssh/id_rsa.pub')
    key = readers.getdataf(keyfile)
    if not key:
        sys.exit(
            'Keyfile {kf} does not exist or not readable'.format(kf=keyfile)
        )
    ssh_init = pxssh()
    ssh_init.SSH_OPTS += (
        " -o 'StrictHostKeyChecking=no'"
        " -o 'CheckHostIP=no'"
        " -o 'UserKnownHostsFile=/dev/null'"
        " -o 'Port={rPORT}'".format(rPORT=remote[1])
    )
    ssh_init.force_password = True
    try:
        ssh_init.login(
            server=remote[0],
            username=cfg.SP_CLUSTER_NAME,
            password=cfg.SP_RAUTH_TOKEN
        )
    except TIMEOUT as err:
        # workaround for a stupid bug in pxssh in centos 6
        if not err.value.startswith('Timeout exceeded in read_nonblocking().'):
            raise
    except (ExceptionPxssh, EOF):
        return False
    pcmd = 'mkdir -p -m700 .ssh && mkdir -p rrsync && echo '
    pcmd += '"command=\\\"rrsync /home/{u}/rrsync\\\" {k}" | '.format(
        k=key,
        u=cfg.SP_CLUSTER_NAME
    )
    pcmd += 'tee -a .ssh/authorized_keys && chmod 600 .ssh/authorized_keys'
    try:
        ssh_init.sendline(pcmd)
        getconf().logger.info(
            'Successfully added %s to %s:%s', key, remote[0], remote[1]
        )
        ssh_init.logout()
    except TIMEOUT as err:
        if not err.value.startswith('Timeout exceeded in read_nonblocking().'):
            raise
    except (ExceptionPxssh, EOF):
        return False
    # test remote connection
    if testremote(remote):
        return True
    return False


def getremotes():
    '''
    Gather the tuples of (remote:port) for all available remotes/ports
    (dies if any of the two is missing or empty)
    returns list of tuples of remotes
    '''
    cfg = getconf()
    addresses = cfg.spconf.get('SP_ABRTSYNC_REMOTE_ADDRESSES', None)
    if not addresses:
        sys.exit(
            (
                'No SP_ABRTSYNC_REMOTE_ADDRESSES configured, at least one '
                'remote address expected, please check'
            )
        )
    addresslist = addresses.split(',')
    ports = cfg.spconf.get('SP_ABRTSYNC_REMOTE_PORTS', None)
    if not ports:
        sys.exit(
            (
                'No SP_ABRTSYNC_REMOTE_PORTS configured, at least one remote '
                'port expected, please check'
            )
        )
    portslist = ports.split(',')
    return [list(i) for i in itertools.product(addresslist, portslist)]


def testremote(remote):
    '''just rsync a test file into remote to check if configured properly
    returns True in case of success, False else'''
    cfg = getconf()
    touch(cfg.testfile)
    tdir = os.path.dirname(cfg.testfile)
    if not sendf(cfg.testfile, tdir, remotes=[remote]):
        return False
    return True


def daemonize():
    '''Used to fork this instance into a daemon'''
    cfg = getconf()
    dpid = os.fork()
    if dpid != 0:
        # pylint: disable=protected-access
        os._exit(0)
    os.setsid()
    dpid = os.fork()
    if dpid != 0:
        # pylint: disable=protected-access
        os._exit(0)
    getconf().logger.info('Changing dir to %s', cfg.basedir)
    os.chdir(cfg.basedir)
    for fdesc in range(0, 1024):
        try:
            os.close(fdesc)
        except OSError:
            pass
    os.open("/dev/null", os.O_RDWR)
    os.dup2(0, 1)
    os.dup2(0, 2)


def ensure_non_blocking_fd(bfd):
    """
    bfd: int
    returns None
    """
    flags = fcntl.fcntl(bfd, fcntl.F_GETFL)
    if flags & os.O_NONBLOCK == 0:
        fcntl.fcntl(bfd, fcntl.F_SETFL, flags | os.O_NONBLOCK)


def lockwait(flockfile):
    '''
    flockfile: file object
    Waits for a minute if another instance is running
    '''
    while True:
        try:
            return fcntl.flock(flockfile, fcntl.LOCK_EX | fcntl.LOCK_NB)
        except IOError as err:
            # raise on unrelated IOErrors
            if err.errno != errno.EAGAIN:
                raise
            else:
                getconf().logger.info(
                    (
                        'Another instance has already obtained lock over '
                        '%s, waiting for 1 minute'
                    ),
                    flockfile
                )
                try:
                    time.sleep(60)
                except KeyboardInterrupt:
                    # just exit silently
                    pass


def update_or_cleanup_sentfiles(flist, bdir, cleanup=None):
    '''flist: list of str file names that were sent
    bdir: str path to base directory for files in flist
    cleanup: boolean to cleanup or not
    if cleanup is True deletes successfully sent files
    Otherwise cleans up files if SP_DELETE_REPORTS is configured,
    otherwise just keeps track of them
    returns None'''
    cfg = getconf()
    if cleanup:
        for ftgt in [os.path.join(bdir, f) for f in flist]:
            cfg.logger.info('Deleting %s', ftgt)
            os.unlink(ftgt)
    else:
        cfg.sentfiles.update([(os.path.join(bdir, i), True) for i in flist])


def getrsshoptslist(cfg, remote):
    """Just constructs rsync ssh command (-e)
    returns str of all required ssh options for rsync"""
    rsshoptslist = [
        "ssh",
        "-o",
        "PasswordAuthentication=no",
        "-o",
        "Port={rPORT}".format(rPORT=remote[1]),
        "-o",
        "User={rUSER}".format(rUSER=cfg.spconf['SP_CLUSTER_NAME']),
        "-o",
        "StrictHostKeyChecking=no",
        "-o",
        "CheckHostIP=no",
        "-o",
        "ConnectTimeout=15s",
        "-o",
        "ControlMaster=auto",
        "-o",
        "ControlPath=~/.ssh/control:%h:%p:%r",
        "-o",
        "ControlPersist=2h",
        "-o",
        "ServerAliveInterval=30"
    ]
    return " ".join(rsshoptslist)


def sendf(files, bdir, remotes=None, cleanup=None):
    '''
    files: str or list of str files to send remotely
    bdir: str base directory to send files from (e.g. /var/crash with files
    /var/crash/dir/file)
    remotes: list of tuple (host, port) if none provided cycles through the
    defaults indefinitely
    cleanup: bool whether to delete files on success

    Validates that each file starts with the base directory, then attempts to
    transfer all files in the list with a single rsync push
    '''
    if remotes is None:
        remotes = itertools.cycle(getremotes())
    if isinstance(files, str):
        files = [files]
    assert all([i.startswith(bdir) for i in files])
    cfg = getconf()
    if cleanup is None:
        if cfg.SP_DELETE_REPORTS:
            cleanup = True
    flist = []
    for filename in files:
        if not os.path.isfile(filename):
            cfg.logger.error("%s does not exist, skipping", filename)
        flist.append(filename)
    fileliststr = '\n'.join(flist) + '\n'
    # only path to the file after the base directory
    if not bdir.endswith('/'):
        # add following '/' to bdir
        bdir += '/'
    filersynclist = '\n'.join([re.sub(bdir, '', i) for i in flist]) + '\n'
    for remote in remotes:
        rsshopts = getrsshoptslist(cfg, remote)
        tfiles = tempfile.NamedTemporaryFile(
            prefix='abrtsync-sendfiles-'
        )
        with open(tfiles.name, 'w') as tff:
            tff.write(filersynclist)
            tff.flush()
            rcmd = [
                'rsync',
                '-e',
                rsshopts,
                '-avc',
                '--partial-dir=.rsync-partial',
                '--files-from={f}'.format(f=tff.name),
                bdir,
                '{rHOST}:{h}/'.format(rHOST=remote[0], h=gethostname())
            ]
            try:
                subprocess.check_call(rcmd)
                cfg.logger.info(
                    'Successfully sent: %s to %s:%s',
                    fileliststr,
                    remote[0],
                    remote[1]
                )
            except subprocess.CalledProcessError:
                cfg.logger.error(
                    'Failed to send through %s, command: %s',
                    remote,
                    ' '.join(rcmd)
                )
                continue
            update_or_cleanup_sentfiles(flist, bdir, cleanup)
            return True
    return False


def ffind(nameglobs, directory):
    '''nameglobs: list of str globs to search for in the present directory
    returns all matching files in a list'''
    cfg = getconf()
    matches = []
    if not directory:
        return matches
    for pattern in nameglobs:
        for root, _, filenames in os.walk(directory):
            if 'lost+found' in root:
                continue
            for filename in fnmatch.filter(filenames, pattern):
                fullpath = os.path.join(root, filename)
                # check that the file was not already sent before
                if cfg.sentfiles.get(fullpath):
                    continue
                matches.append(fullpath)
    return matches


def cleanrepsync():
    '''
    cleanup old repsync directory
    '''
    def rmfiles(directory, filelist):
        '''Removes filelist from directory, returns None'''
        files = [os.path.join(directory, i) for i in filelist if i != '']
        for ofl in files:
            try:
                os.remove(ofl)
                dirname = os.path.dirname(ofl)
                if dirname:
                    os.rmdir(dirname)
            except OSError:
                # file/dir does not exist
                pass

    try:
        pwrec = pwd.getpwnam(getconf().spconf.get('SP_ABRTSYNC_USER', None))
    except (KeyError, TypeError):
        getconf().logger.info(
            'no SP_ABRTSYNC_USER or user missing, nothing to cleanup'
        )
        return
    repsyncdir = os.path.join(pwrec.pw_dir, 'repsync')
    if os.path.isdir(repsyncdir):
        getconf().logger.info('%s exists, proceeding with cleanup', repsyncdir)
        # check for old reports and cleanup
        rfile = os.path.join(repsyncdir, 'reports')
        if os.path.isfile(rfile):
            files = getdataflines(rfile)
            if files:
                getconf().logger.info(
                    '%s exists, cleaning up files: %s', rfile, '\n'.join(files)
                )
                rmfiles(repsyncdir, files)
        vfile = os.path.join(repsyncdir, 'vmcores')
        if os.path.isfile(vfile):
            vmcorefiles = getdataflines(vfile)
            if vmcorefiles:
                varcrashdir = getvarcrash()
                if os.path.isdir(varcrashdir):
                    getconf().logger.info(
                        '%s exists, cleaning up files: %s',
                        rfile,
                        '\n'.join(files)
                    )
                    vfiles = [
                        os.path.join('/var/crash/', i) for i in vmcorefiles
                        if i != ''
                    ]
                    rmfiles(varcrashdir, vfiles)
        getconf().logger.info("Removing %s directory", repsyncdir)
        shutil.rmtree(repsyncdir)


def getconf():
    '''ensures that the STATCONFIG instance exists
    returns the AbrtsyncConfig instance object'''
    if 'STATCONFIG' not in dir(getconf):
        getconf.STATCONFIG = AbrtsyncConfig()
    return getconf.STATCONFIG


def _main():
    '''initial configuration and main loop'''
    getconf().logger = get_logger()
    cfg = getconf()
    cfg.args = getargs()
    writers.write_file(cfg.args.pidfile, str(cfg.ppid))
    init()
    if cfg.args.watch_fd is not None:
        ensure_non_blocking_fd(cfg.args.watch_fd)
    reportdir = cfg.spconf['SP_REPORTDIR']
    cleanrepsync()
    if cfg.args.daemon:
        daemonize()
    global_lock = '{b}/{s}.lock'.format(
        b=cfg.basedir,
        s=os.path.basename(sys.argv[0])
    )
    with open(global_lock, 'w+') as lock:
        lockwait(lock)
        for remote in getremotes():
            # configure a remote on start
            if not configureremote(remote):
                cfg.logger.info('Failed to configure %s', remote)
        # check for files in VARCRASH directory for sending
        kdumps = ffind(['*dmesg*'], cfg.VARCRASH)
        if kdumps:
            sendf(kdumps, cfg.VARCRASH, cleanup=False)
        while True:
            newreports = ffind(['report-*', '*.json'], reportdir)
            if newreports:
                sendf(newreports, reportdir)
            if cfg.args.noop:
                cfg.logger.info('-N/--noop selected, exiting.')
                exit(0)
            time.sleep(0.5)


def main():
    '''main function with exception logging'''
    try:
        _main()
    except KeyboardInterrupt:
        getconf().logger.info('Bye')
    except Exception:   # pylint: disable=broad-except
        stack = ''.join(traceback.format_stack())
        getconf().logger.exception(stack)
        sys.exit(stack)


if __name__ == "__main__":
    main()
