"""Provides different function decorators."""

import errno
import functools
import os
import signal
import time
try:
    from decorator import decorator
except ImportError:
    # pylint: disable=missing-docstring
    def decorator(caller):
        def decor(func):
            @functools.wraps(func)
            def wrapper(*args, **kwargs):
                return caller(func, *args, **kwargs)
            return wrapper
        return decor


def static_vars(**kwargs):
    """Adds static variables to the function and initialize them with the given values."""
    def _decorate(func):
        for k in kwargs:
            setattr(func, k, kwargs[k])
        return func

    return _decorate


class TimeoutError(Exception):
    """Timeout exception for the timeout decorator."""
    pass


def timeout(seconds, error_message=os.strerror(errno.ETIME)):
    """
    Timeout decorator. Utilizes SIGALRM to stop a function after given amount
    of seconds if it hasn't finished.
    """
    def _decorator(func):
        def _handle_timeout(signum, frame):
            _sig_info = (signum, frame)
            raise TimeoutError(error_message)

        def wrapper(*args, **kwargs):
            # pylint: disable=missing-docstring
            signal.signal(signal.SIGALRM, _handle_timeout)
            signal.alarm(seconds)
            try:
                result = func(*args, **kwargs)
            finally:
                signal.alarm(0)
            return result

        return functools.wraps(func)(wrapper)

    return _decorator


def retry(exceptions=None, tries=-1, delay=0, backoff=1, logger=None):
    """Returns a retry decorator.
    exceptions: Exception or tuple of exceptions to catch. default: Exception
    tries: the maximum number of attempts. default: -1 (infinite).
    delay: initial delay between attempts. default: 0.
    backoff: modifier multiplied to delay between attempts.
    default: 1 (no backoff).
    returns retry decorator
    """

    def _retry_internal(func, exceptions, _tries, _delay, backoff):
        """
        Executes a function and retries it if it failed.
        func: the function to execute.
        """
        while _tries:
            try:
                return func()
            except exceptions as err:
                _tries -= 1
                if not _tries:
                    raise
                if logger:
                    logger.info('%s, retrying in %s seconds...', err, _delay)
                time.sleep(_delay)
                _delay *= backoff

    @decorator
    def retry_decorator(func, *args, **kwargs):
        # pylint: disable=missing-docstring
        return _retry_internal(
            functools.partial(func, *args, **kwargs),
            exceptions,
            tries,
            delay,
            backoff,
        )
    return retry_decorator
